import os
import numpy as np
import contextlib
import mmap
import numpy.distutils.system_info as sysinfo


def get_active_tetrode(set_filename):
    """in the .set files it will say collectMask_X Y for each tetrode number to tell you if
    it is active or not. T1 = ch1-ch4, T2 = ch5-ch8, etc."""
    active_tetrode = []
    active_tetrode_str = 'collectMask_'

    with open(set_filename, encoding='cp1252') as f:
        for line in f:

            # collectMask_X Y, where x is the tetrode number, and Y is eitehr on or off (1 or 0)
            if active_tetrode_str in line:
                tetrode_str, tetrode_status = line.split(' ')
                if int(tetrode_status) == 1:
                    # then the tetrode is saved
                    tetrode_str.find('_')
                    tet_number = int(tetrode_str[tetrode_str.find('_') + 1:])
                    active_tetrode.append(tet_number)

    return active_tetrode


def get_active_eeg(set_filename):
    """This will return a dictionary (cative_eeg_dict) where the keys
    will be eeg channels from 1->64 which will represent the eeg suffixes (2 = .eeg2, 3 = 2.eeg3, etc)
    and the key will be the channel that the EEG maps to (a channel from 0->63)"""
    active_eeg = []
    active_eeg_str = 'saveEEG_ch'

    eeg_map = []
    eeg_map_str = 'EEG_ch_'

    active_eeg_dict = {}

    with open(set_filename, encoding='cp1252') as f:
        for line in f:

            if active_eeg_str in line:
                # saveEEG_ch_X Y, where x is the eeg number, and Y is eitehr on or off (1 or 0)
                _, status = line.split(' ')
                active_eeg.append(int(status))
            elif eeg_map_str in line:
                # EEG_ch_X Y
                _, chan = line.split(' ')
                eeg_map.append(int(chan))

                # active_eeg = np.asarray(active_eeg)
                # eeg_map = np.asarray(eeg_map)

    for i, status in enumerate(active_eeg):
        if status == 1:
            active_eeg_dict[i + 1] = eeg_map[i] - 1

    return active_eeg_dict


def get_bin_data(bin_filename, channels=None, tetrode=None):
    """This function will be used to acquire the actual lfp data given the .bin filename,
    and the tetrode or channels (from 1-64) that you want to get"""

    if tetrode is not None:
        channels = get_channel_from_tetrode(tetrode)
    else:
        channels = np.array(channels)  # just in case it isn't an np.array

    bytes_per_iteration = 432

    with open(bin_filename, 'rb') as f:
        # pass
        with contextlib.closing(mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ)) as m:
            num_iterations = int(len(m)/bytes_per_iteration)

            data = np.ndarray((num_iterations,), (np.int16, (1,192)), m, 32, (bytes_per_iteration,)).reshape((-1, 1)).flatten()
            data = samples_to_array(data, channels=channels.tolist())

    return data


def samples_to_array(A, channels=[]):
    """This will take data matrix A, and convert it into a numpy array, there are three samples of
    64 channels in this matrix, however their channels do need to be re-mapped"""

    if channels == []:
        channels = np.arange(64) + 1
    else:
        channels = np.asarray(channels)

    A = np.asarray(A)

    sample_num = int(len(A) / 64)  # get the sample numbers

    sample_array = np.zeros((len(channels), sample_num))  # creating a 64x3 array of zeros (64 channels, 3 samples)

    for i, channel in enumerate(channels):
        sample_array[i, :] = A[get_sample_indices(channel, sample_num)]

    return sample_array


def get_sample_indices(channel_number, samples):
    remap_channel = get_remap_chan(channel_number)

    indices_scalar = np.multiply(np.arange(samples), 64)
    sample_indices = indices_scalar + np.multiply(np.ones(samples), remap_channel)

    # return np.array([remap_channel, 64 + remap_channel, 64*2 + remap_channel])
    return (indices_scalar + np.multiply(np.ones(samples), remap_channel)).astype(int)


def get_remap_chan(chan_num):
    """There is re-mapping, thus to get the correct channel data, you need to incorporate re-mapping
    input will be a channel from 1 to 64, and will return the remapped channel"""

    remap_channels = np.array([32, 33, 34, 35, 36, 37, 38, 39, 0, 1, 2, 3, 4, 5,
                               6, 7, 40, 41, 42, 43, 44, 45, 46, 47, 8, 9, 10, 11,
                               12, 13, 14, 15, 48, 49, 50, 51, 52, 53, 54, 55, 16, 17,
                               18, 19, 20, 21, 22, 23, 56, 57, 58, 59, 60, 61, 62, 63,
                               24, 25, 26, 27, 28, 29, 30, 31])

    return remap_channels[chan_num - 1]


def get_channel_from_tetrode(tetrode):
    """This function will take the tetrode number and return the Axona channel numbers
    i.e. Tetrode 1 = Ch1 -Ch4, Tetrode 2 = Ch5-Ch8, etc"""
    tetrode = int(tetrode)  # just in case the user gave a string as the tetrode

    return np.arange(1, 5) + 4 * (tetrode - 1)


def get_raw_pos(bin_filename, mode='mmap'):
    """This will get the raw position data from the .bin file in the following format:
    video timestamp, x1, y1, x2, y2, numpix1, numpix2, total_pix, unused value

    The video timestamp is the time since the camera has been turned on, thus this value is
    irrelevent

    The .bin file is written of packets of 432 bytes containing position data, three samples of
    electrophys data (per channel), and etc.

    A packet only has a valid position if it has an "ADU2" tag in the header. You'll notice that
    the positions are sampled at twice the normal rate to avoid aliasing.

    There are two modes, one will use memory mapping, and the other will not: 'mmap' or 'non'
    """
    # pos_sample_num = 0

    byte_count = os.path.getsize(bin_filename)
    bytes_per_iteration = 432
    iteration_count = int(byte_count / bytes_per_iteration)
    # sample_count = iteration_count * 192  # each iteration has 192 samples (64*3)

    optimal_iteration = 1000000
    if optimal_iteration >= iteration_count:
        simul_iterations = iteration_count
    else:
        simul_iterations = find_n(iteration_count, optimal=optimal_iteration)

    n = int(iteration_count / simul_iterations)  # finds how many loops to do
    byte_chunksize = int(simul_iterations * bytes_per_iteration)

    DaqFs = 48000
    duration = iteration_count * 3 / DaqFs
    duration = np.ceil(duration)

    pos_Fs = 50
    n_samples = int(duration * pos_Fs)

    # Reading the Data

    # header_byte_len = 32
    # data_byte_len = 384
    # trailer_byte_len = 16

    raw_pos = np.array([]).astype(float)

    if sysinfo.platform_bits != 64:
        mode = 'non'  # do not mmap with a 32 bit python

    with open(bin_filename, 'rb') as f:

        if mode == 'mmap':
            # we will use memory mapped objects

            with contextlib.closing(mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ)) as m:
                num_iterations = int(len(m) / bytes_per_iteration)

                byte_ids = np.ndarray((num_iterations,), 'S4', m, 0, bytes_per_iteration)

                pos_bool = np.where(byte_ids == b'ADU2')  # valid position packets have an ADU2 header b'ADU2' in bytes

                valid_iterations = np.arange(len(byte_ids))[pos_bool]  # getting the packet number

                # timestamp starts at the 12th bit from the start
                time_stamp = np.ndarray((num_iterations,), np.int32, m, 12, bytes_per_iteration)[pos_bool].reshape(
                    (-1, 1))  # gettinge the time stamp data

                # position values start from the 16th bit
                raw_pos = np.ndarray((num_iterations,), (np.int16, (1, 8)), m, 16,
                                     (bytes_per_iteration,)).reshape((-1, 8))[pos_bool][:]

                raw_pos = np.hstack((valid_iterations.reshape((-1, 1)), time_stamp, raw_pos)).astype(
                    float)  # stacking all these values to create one matrix

        else:

            # in some cases using memory mapping will be slow, such as using 32 bit python

            iteration_start = 0
            # we will iterate
            for i in range(n):
                data = f.read(byte_chunksize)
                num_iterations = int(len(data) / bytes_per_iteration)

                byte_ids = np.ndarray((num_iterations,), 'S4', data, 0, 432)

                pos_bool = np.where(byte_ids == b'ADU2')  # valid position packets have an ADU2 header b'ADU2' in bytes

                valid_iterations = np.arange(len(byte_ids))[pos_bool]  # getting the packet number

                time_stamp = np.ndarray((num_iterations,), np.uint32, data, 12, 432)[pos_bool].reshape(
                    (-1, 1))  # gettinge the time stamp data

                positions = np.ndarray((num_iterations,), (np.uint16, (1, 8)), data, 16, (432,)).reshape((-1, 8))[
                                pos_bool][:]

                i = np.add(valid_iterations,
                           iteration_start)  # offsetting the samples depending on the chunk number (n)

                positions = np.hstack((i.reshape((-1, 1)), time_stamp, positions)).astype(
                    float)  # stacking all these values to create one matrix

                iteration_start += simul_iterations

                if len(raw_pos) != 0:

                    raw_pos = np.vstack((raw_pos, positions))

                else:

                    raw_pos = positions

        # raw_pos[:, 1] = np.divide(raw_pos[:, 1], 50)  # converting from a frame count to a time value in seconds

        # raw_pos is structured as: packet #, video timestamp, y1, x1, y2, x2, numpix1, numpix2, total_pix, unused value

        # the X and Y values are reverse piece-wise so lets switch the format from
        # packet #, video timestamp, y1, x1, y2, x2, numpix1, numpix2, total_pix, unused value
        # to
        # packet #, video timestamp, x1, y1, x2, y2, numpix1, numpix2, total_pix, unused value

        raw_pos[:, 2:6] = raw_pos[:, [3, 2, 5, 4]]

        # find the first valid sample, since the data is sampled at 48kHz, and there are 3 samples per packet, the packet
        # rate is 16kHz. The positions are sampled at 50 Hz thus there is a valid position ever 320 packets. The valid position
        # will essentially take the last ADU2 headered packet values

        first_sample_index = len(
            np.where(raw_pos[:, 0] <= 320 - 1)[0]) - 1  # subtract one since indices in python start at 0

        raw_pos = raw_pos[first_sample_index:, :]

        # there should be twice the number of samples since they double sampled to stop aliasing
        # if there is not 2 * n_samples, append the last recorded sample to the end (we will assume the animal remained there
        # for the rest of the time)

        if raw_pos.shape[0] < 2 * n_samples:
            missing_n = int(2 * n_samples - raw_pos.shape[0])
            last_location = raw_pos[-1, :]
            missing_samples = np.tile(last_location, (missing_n, 1))

            raw_pos = np.vstack((raw_pos, missing_samples))

        # now we will set the oversampled data to 1023 (NaN) as that is how the converter treats the double sampled data

        # indices = np.arange(1,raw_pos.shape[0], 2)
        indices = np.arange(0, raw_pos.shape[0], 2)
        if indices[-1] >= raw_pos.shape[0]:
            indices = indices[:-1]

        # raw_pos[indices, 2:4] = 1023
        raw_pos = raw_pos[indices, :]

        raw_pos = raw_pos[:, 1:]  # don't need the packet index anymore

    return raw_pos


def find_n(iterations, optimal=1000):
    n = optimal

    while True:
        if iterations % n == 0:

            return n
        else:
            n -= 1
    return 'abort'



import os
from .readBin import get_active_tetrode, get_active_eeg


def get_set_header(set_filename):
    with open(set_filename, 'r+') as f:
        header = ''
        for line in f:
            header += line
            if 'sw_version' in line:
                break
    return header


def is_egf_active(set_filename):
    active_egf_str = 'saveEGF'

    with open(set_filename) as f:
        for line in f:

            if active_egf_str in line:
                _, egf_status = line.split(' ')

                if int(egf_status) == 1:
                    return True

        return False


def is_converted(set_filename):
    """This method will check if the file has been converted already"""

    # find active tetrodes

    active_tetrodes = get_active_tetrode(set_filename)

    # check if each of these active .N files have been created yet

    set_basename = os.path.basename(os.path.splitext(set_filename)[0])

    set_directory = os.path.dirname(set_filename)

    all_tetrodes_written = True
    all_eeg_written = True
    all_egf_written = True
    pos_written = True

    if not os.path.exists(os.path.join(set_directory, '%s.pos' % (set_basename))):
        pos_written = False

    for tetrode in active_tetrodes:

        if not os.path.exists(os.path.join(set_directory, '%s.%d' % (set_basename, int(tetrode)))):
            all_tetrodes_written = False
            break

    active_eeg = get_active_eeg(set_filename)

    for eeg in active_eeg.keys():
        if eeg == 1:
            if not os.path.exists(os.path.join(set_directory, '%s.eeg' % (set_basename))):
                all_eeg_written = False
                break

        elif not os.path.exists(os.path.join(set_directory, '%s.eeg%d' % (set_basename, int(eeg)))):
            all_eeg_written = False
            break

    if is_egf_active(set_filename):
        for egf in active_eeg.keys():
            if egf == 1:
                if not os.path.exists(os.path.join(set_directory, '%s.egf' % (set_basename))):
                    all_egf_written = False
                    break

            elif not os.path.exists(os.path.join(set_directory, '%s.egf%d' % (set_basename, int(egf)))):
                all_egf_written = False
                break

    return all_tetrodes_written and all_eeg_written and all_egf_written and pos_written


def has_files(set_filename):
    """This method will check if all the necessary files exist"""

    # the only thing it needs is a .bin file

    tint_basename = os.path.basename(os.path.splitext(set_filename)[0])

    directory = os.path.dirname(set_filename)

    if os.path.exists(os.path.join(directory, '%s.bin' % tint_basename)):
        return True

    return False



import os
from .conversion_utils import get_set_header
import numpy as np
import struct


def write_tetrode(filepath, data, Fs):

    session_path, session_filename = os.path.split(filepath)
    tint_basename = os.path.splitext(session_filename)[0]
    set_filename = os.path.join(session_path, '%s.set' % tint_basename)

    n = len(data)

    header = get_set_header(set_filename)

    with open(filepath, 'w') as f:
        num_chans = 'num_chans 4'
        timebase_head = '\ntimebase %d hz' % (96000)
        bp_timestamp = '\nbytes_per_timestamp %d' % (4)
        # samps_per_spike = '\nsamples_per_spike %d' % (int(Fs*1e-3))
        samps_per_spike = '\nsamples_per_spike %d' % (50)
        sample_rate = '\nsample_rate %d hz' % (Fs)
        b_p_sample = '\nbytes_per_sample %d' % (1)
        # b_p_sample = '\nbytes_per_sample %d' % (4)
        spike_form = '\nspike_format t,ch1,t,ch2,t,ch3,t,ch4'
        num_spikes = '\nnum_spikes %d' % (n)
        start = '\ndata_start'

        write_order = [header, num_chans, timebase_head,
                       bp_timestamp,
                       samps_per_spike, sample_rate, b_p_sample, spike_form, num_spikes, start]

        f.writelines(write_order)

    # rearranging the data to have a flat array of t1, waveform1, t2, waveform2, t3, waveform3, etc....
    spike_times = np.asarray(sorted(data.keys()))

    # the spike times are repeated for each channel so lets tile this
    spike_times = np.tile(spike_times, (4, 1))
    spike_times = spike_times.flatten(order='F')

    spike_values = np.asarray([value for (key, value) in sorted(data.items())])

    # this will create a (n_samples, n_channels, n_samples_per_spike) => (n, 4, 50) sized matrix, we will create a
    # matrix of all the samples and channels going from ch1 -> ch4 for each spike time
    # time1 ch1_data
    # time1 ch2_data
    # time1 ch3_data
    # time1 ch4_data
    # time2 ch1_data
    # time2 ch2_data
    # .
    # .
    # .

    spike_values = spike_values.reshape((n * 4, 50))  # create the 4nx50 channel data matrix

    # make the first column the time values
    spike_array = np.hstack((spike_times.reshape(len(spike_times), 1), spike_values))

    data = None
    spike_times = None
    spike_values = None

    spike_n = spike_array.shape[0]

    t_packed = struct.pack('>%di' % spike_n, *spike_array[:, 0].astype(int))
    spike_array = spike_array[:, 1:]  # removing time data from this matrix to save memory

    spike_data_pack = struct.pack('<%db' % (spike_n*50), *spike_array.astype(int).flatten())

    spike_array = None

    # now we need to combine the lists by alternating

    comb_list = [None] * (2*spike_n)
    comb_list[::2] = [t_packed[i:i + 4] for i in range(0, len(t_packed), 4)]  # breaks up t_packed into a list,
    # each timestamp is one 4 byte integer
    comb_list[1::2] = [spike_data_pack[i:i + 50] for i in range(0, len(spike_data_pack), 50)]  # breaks up spike_data_
    # pack and puts it into a list, each spike is 50 one byte integers

    t_packed = None
    spike_data_pack = None

    write_order = []
    with open(filepath, 'rb+') as f:

        write_order.extend(comb_list)
        write_order.append(bytes('\r\ndata_end\r\n', 'utf-8'))

        f.seek(0, 2)
        f.writelines(write_order)



import os
from PyQt5 import QtWidgets
import time
import threading
from .conversion_utils import has_files, is_converted


threadLock = threading.Lock()


def addSessions(self):
    """Adds any sessions that are not already on the list"""

    if self is not None:
        # doesn't add sessions when re-ordering
        while self.modifying_list:
            # pauses add Sessions when the individual is reordering
            time.sleep(0.1)

    directory_added = False

    current_directory = self.directory_edit.text()
    # finds the sub directories within the chosen directory
    try:
        sub_directories = [d for d in os.listdir(current_directory)
                           if os.path.isdir(os.path.join(current_directory, d)) and
                           len([file for file in os.listdir(os.path.join(current_directory, d))
                                if '.set' in file]) != 0 and
                               d not in ['Processed', 'Converted']]
    except OSError:
        return

    iterator = QtWidgets.QTreeWidgetItemIterator(self.recording_queue)
    # loops through all the already added sessions
    added_directories = []
    while iterator.value():
        item = iterator.value()

        check_directory = os.path.join(current_directory, item.data(0, 0))

        if not os.path.exists(check_directory) and '.set' not in item.data(0, 0):
            # then remove from the list since it doesn't exist anymore
            while threadLock:
                if os.path.basename(check_directory) != self.current_subdirectory:
                    root = self.recording_queue.invisibleRootItem()
                    for child_index in range(root.childCount()):
                        if root.child(child_index) == item:
                            self.RemoveChildItem.myGUI_signal_QTreeWidgetItem.emit(item)
        else:
            added_directories.append(item.data(0, 0))

        iterator += 1

    for directory in sub_directories:

        if directory in added_directories:
            # the directory has already been added, determine if we need to add more
            # find the treewidget item
            iterator = QtWidgets.QTreeWidgetItemIterator(self.recording_queue)
            while iterator.value():
                directory_item = iterator.value()
                if directory_item.data(0, 0) == directory:
                    break
                iterator += 1

            # find added sessions
            added_sessions = []
            try:
                iterator = QtWidgets.QTreeWidgetItemIterator(directory_item)
            except UnboundLocalError:
                return
            except RuntimeError:
                return

            # compile list of added sessions
            while iterator.value():
                session_item = iterator.value()
                added_sessions.append(session_item.data(0, 0))
                iterator += 1

            # find sessions to add
            try:
                sessions = FindSessions(os.path.join(current_directory, directory))
            except FileNotFoundError:
                return
            except PermissionError:
                return

            for session in sessions:
                if session not in added_sessions:
                    tint_basename = os.path.basename(session)

                    # only adds the sessions that haven't been added already

                    session_item = QtWidgets.QTreeWidgetItem()
                    session_item.setText(0, tint_basename)

                    directory_item.addChild(session_item)
                else:
                    pass

        else:
            directory_item = QtWidgets.QTreeWidgetItem()
            directory_item.setText(0, directory)

            # find sessions to add
            try:
                sessions = FindSessions(os.path.join(current_directory, directory))
            except FileNotFoundError:
                return
            except PermissionError:
                return

            # add the sessions to the TreeWidget
            for session in sessions:
                tint_basename = os.path.basename(session)

                # only adds the sessions that haven't been added already

                session_item = QtWidgets.QTreeWidgetItem()
                session_item.setText(0, tint_basename)

                directory_item.addChild(session_item)

            if directory_item.childCount() != 0:
                # makes sure that it only adds sessions that have sessions to convert
                self.recording_queue.addTopLevelItem(directory_item)

                directory_added = True

    if directory_added:
        pass


def FindSessions(directory):
    """This function will find the sessions"""

    directory_file_list = os.listdir(
        directory)  # making a list of all files within the specified directory

    set_filenames = []

    [set_filenames.append(file) for file in directory_file_list if
     '.set' in file and has_files(os.path.join(directory, file)) and not
     is_converted(os.path.join(directory, file))]

    return set_filenames


def RepeatAddSessions(main_window):
    """
    This will repeat adding the sessions for BatchTINTV3 so that it is continuously looking for files to analyze within
    a chosen directory.

    :param main_window: this is again the self of the main window for batchTINTV3
    :return:
    """

    # the thread is active, set to true

    main_window.repeat_thread_active = True

    while True:
        with threadLock:
            if main_window.reset_add_thread:
                main_window.repeat_thread_active = False
                main_window.reset_add_thread = False
                return

        if main_window.directory_changed:
            # then we have changed the append cut value
            while (time.time() - main_window.change_directory_time) < 0.5:
                time.sleep(0.1)
            main_window.recording_queue.clear()
            main_window.directory_changed = False

        try:
            with threadLock:
                main_window.adding_session = True
            addSessions(main_window)

            with threadLock:
                main_window.adding_session = False
                time.sleep(0.1)
        except FileNotFoundError:
            pass
        except RuntimeError:
            pass




import sys, shutil, os, time, datetime
from PyQt5 import QtCore, QtGui, QtWidgets
from core.ConversionFunctions import convert_basename
from distutils.dir_util import copy_tree
from BatchTINTV3.core.settings import Settings_Window
from BatchTINTV3.core.klusta_functions import klusta
from core.defaultParameters import default_batchtint, default_filename, default_move_converted, default_threshold
from core.utils import background, Worker, center, Communicate, project_name, raise_w
from core.AddSessions import RepeatAddSessions
import json


class Window(QtWidgets.QWidget):  # defines the window class (main window)

    def __init__(self):  # initializes the main window
        super(Window, self).__init__()
        background(self)  # acquires some features from the background function we defined earlier
        # sets the title of the window

        self.setWindowTitle("%s - Main Window" % project_name)

        # self.current_session = None
        self.current_subdirectory = None
        self.directory_changed = False
        self.modifying_list = False
        self.reset_add_thread = False
        self.repeat_thread_active = True
        self.conversion = False
        self.choice = None
        self.file_chosen = False

        # this signal will append errors/messages to the Log object
        self.LogAppend = Communicate()
        self.LogAppend.myGUI_signal.connect(self.AppendLog)

        # this signal will raise any errors
        self.LogError = Communicate()
        self.LogError.myGUI_signal.connect(self.raiseError)

        # this will remove the top level item from the queue
        self.RemoveQueueItem = Communicate()
        self.RemoveQueueItem.myGUI_signal.connect(self.takeTopLevel)

        self.RemoveSessionItem = Communicate()
        self.RemoveSessionItem.myGUI_signal.connect(self.takeChild)

        self.RemoveSessionData = Communicate()
        self.RemoveSessionData.myGUI_signal.connect(self.takeChildData)

        self.SetSessionItem = Communicate()
        self.SetSessionItem.myGUI_signal.connect(self.setChild)

        self.RemoveChildItem = Communicate()
        self.RemoveChildItem.myGUI_signal_QTreeWidgetItem.connect(self.removeChild)

        self.RepeatAddSessionsThread = QtCore.QThread()
        self.convert_thread = QtCore.QThread()

        self.home()  # runs the home function

    def home(self):  # defines the home function (the main window)

        # read any previous directories chosen so we don't have to
        try:  # attempts to open previous directory catches error if file not found
            # No saved directory's need to create file
            with open(self.directory_settings, 'r+') as filename:  # opens the defined file
                directory_data = json.load(filename)  # loads the directory data from file
                if os.path.exists(directory_data['directory']):
                    current_directory_name = directory_data['directory']  # defines the data
                else:
                    current_directory_name = default_filename  # states that no directory was chosen

        except FileNotFoundError:  # runs if file not found
            with open(self.directory_settings, 'w') as filename:  # opens a file
                current_directory_name = default_filename  # states that no directory was chosen
                directory_data = {'directory': current_directory_name}  # creates a dictionary
                json.dump(directory_data, filename)  # writes the dictionary to the file

        # ------ buttons + widgets -----------------------------

        quit_btn = QtWidgets.QPushButton("Quit", self)
        quit_btn.clicked.connect(self.close_app)
        quit_btn.setShortcut("Ctrl+Q")
        quit_btn.setToolTip('Click to quit (or press Ctrl+Q)')

        self.convert_button = QtWidgets.QPushButton('Convert', self)
        self.convert_button.clicked.connect(self.Convert)
        self.convert_button.setToolTip('Click to start the conversion.')

        self.batch_tint_settings_window = None
        self.batch_tint_settings_button = QtWidgets.QPushButton("Batch Tint Settings")
        self.batch_tint_settings_button.clicked.connect(self.open_batch_tint_settings)

        btn_layout = QtWidgets.QHBoxLayout()

        button_order = [self.convert_button, self.batch_tint_settings_button, quit_btn]
        for button in button_order:
            btn_layout.addWidget(button)

        # Version information -------------------------------------------
        vers_label = QtWidgets.QLabel("%s V1.0.3" % (project_name, ))

        # ------------------ widget layouts ----------------
        self.choose_directory_btn = QtWidgets.QPushButton('Choose Directory', self)
        self.choose_directory_btn.clicked.connect(self.new_directory)

        # the label that states that the line-edit corresponds to the current directory
        directory_label = QtWidgets.QLabel('Current Directory')
        directory_label.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)

        # the line-edit that displays the current directory
        self.directory_edit = QtWidgets.QLineEdit()
        self.directory_edit.setAlignment(QtCore.Qt.AlignCenter | QtCore.Qt.AlignVCenter)  # aligning the text
        self.directory_edit.setText(current_directory_name)  # default text
        # updates the directory every time the text changes
        self.directory_edit.textChanged.connect(self.changed_directory)

        self.batch_tint_checkbox = QtWidgets.QCheckBox("Batch Tint")
        self.batch_tint_checkbox.toggle()

        self.move_converted_checkbox = QtWidgets.QCheckBox("Move Converted")
        self.move_converted_checkbox.toggle()

        # creating the layout for the text + line-edit so that they are aligned appropriately
        current_directory_layout = QtWidgets.QHBoxLayout()
        current_directory_layout.addWidget(directory_label)
        current_directory_layout.addWidget(self.directory_edit)
        current_directory_layout.addWidget(self.batch_tint_checkbox)
        current_directory_layout.addWidget(self.move_converted_checkbox)

        # creating a layout with the line-edit/text + the button so that they are all together
        directory_layout = QtWidgets.QHBoxLayout()
        directory_layout.addWidget(self.choose_directory_btn)
        directory_layout.addLayout(current_directory_layout)

        # creates the queue of recording sessions to convert
        self.recording_queue = QtWidgets.QTreeWidget()
        self.recording_queue.headerItem().setText(0, "Recording Session:")
        recording_queue_label = QtWidgets.QLabel("Conversion Queue:")
        recording_queue_label.setFont(QtGui.QFont("Arial", 10, weight=QtGui.QFont.Bold))

        recording_queue_layout = QtWidgets.QVBoxLayout()
        recording_queue_layout.addWidget(recording_queue_label)
        recording_queue_layout.addWidget(self.recording_queue)

        # adding the layout for the log
        self.log = QtWidgets.QTextEdit()
        log_label = QtWidgets.QLabel('Log:')
        log_label.setFont(QtGui.QFont("Arial", 10, weight=QtGui.QFont.Bold))
        log_layout = QtWidgets.QVBoxLayout()
        log_layout.addWidget(log_label)
        log_layout.addWidget(self.log)

        # adding the thresholding portion fo the layout

        threshold_label = QtWidgets.QLabel('Threshold(SD\'s)')
        self.threshold = QtWidgets.QLineEdit()
        self.threshold.setAlignment(QtCore.Qt.AlignCenter | QtCore.Qt.AlignVCenter)
        self.threshold.setText(str(default_threshold))
        self.threshold.setToolTip("This will determine the standard deviations away from the baseline value " +
                                  "that you want to use for the thresholding")

        # check if you want to perform a DC block on the EEG and EGF data
        self.dc_blocker = QtWidgets.QCheckBox("DC Blocking Filter")
        self.dc_blocker.toggle()  # set the default to on

        threshold_layout = QtWidgets.QHBoxLayout()
        for widget in [threshold_label, self.threshold]:
            threshold_layout.addWidget(widget)

        # extra parameters layout

        parameters_layout = QtWidgets.QHBoxLayout()
        for parameter in [threshold_layout, self.dc_blocker]:
            if 'Layout' in parameter.__str__():
                parameters_layout.addLayout(parameter)
            else:
                parameters_layout.addWidget(parameter)

        # ------------- layout ------------------------------

        layout = QtWidgets.QVBoxLayout()

        layout_order = [directory_layout, recording_queue_layout, log_layout,
                        parameters_layout, btn_layout]

        for order in layout_order:
            if 'Layout' in order.__str__():
                layout.addLayout(order)
                layout.addStretch(1)
            else:
                layout.addWidget(order, 0, QtCore.Qt.AlignCenter)
                layout.addStretch(1)

        layout.addStretch(1)  # adds stretch to put the version info at the button
        layout.addWidget(vers_label)  # adds the date modification/version number

        self.setSettings()

        self.setLayout(layout)

        center(self)

        self.set_batch_tint_settings()

        self.show()

        # start thread that will search for new files to convert
        self.RepeatAddSessionsThread.start()
        self.RepeatAddSessionsWorker = Worker(RepeatAddSessions, self)
        self.RepeatAddSessionsWorker.moveToThread(self.RepeatAddSessionsThread)
        self.RepeatAddSessionsWorker.start.emit("start")

    def setCheckbox(self, widget, value):

        if widget.isChecked() and not value:
            # if the checkbox is checked and it isn't supposed to be, untoggle
            widget.toggle()
        elif not widget.isChecked and value:
            # if checkbox isn't checked, and it's supposed to be, toggle
            widget.toggle()

    def setValues(self, settings):
        self.setCheckbox(self.batch_tint_checkbox, settings['batchtint'])
        self.setCheckbox(self.move_converted_checkbox, settings['move_converted'])
        self.threshold.setText(str(settings['threshold']))

    def setSettings(self):
        try:
            with open(self.settings_fname, 'r+') as filename:
                settings = json.load(filename)
        except FileNotFoundError:
            self.setDefaultSettings()
            with open(self.settings_fname, 'r+') as filename:
                settings = json.load(filename)

        try:
            self.setValues(settings)
        except KeyError:
            # re-create settings file, likely you have an outdated version
            self.setDefaultSettings()
            self.setValues(settings)

    def setDefaultSettings(self):
        settings = {}

        settings['batchtint'] = default_batchtint
        settings['move_converted'] = default_move_converted
        settings['threshold'] = default_threshold

        with open(self.settings_fname, 'w') as filename:
            json.dump(settings, filename)

    def getSettings(self):
        settings = {}

        settings['batchtint'] = self.batch_tint_checkbox.isChecked()
        settings['move_converted'] = self.move_converted_checkbox.isChecked()
        settings['threshold'] = self.threshold.text()

        return settings

    def restart_add_sessions_thread(self):

        self.reset_add_thread = True
        if not hasattr(self, 'repeat_thread_active'):
            return

        while self.repeat_thread_active:
            time.sleep(0.1)

        self.RepeatAddSessionsThread.setTerminationEnabled(True)
        self.RepeatAddSessionsThread.start()

        self.RepeatAddSessionsWorker = Worker(RepeatAddSessions, self)
        self.RepeatAddSessionsWorker.moveToThread(self.RepeatAddSessionsThread)
        self.RepeatAddSessionsWorker.start.emit("start")

    def set_batch_tint_settings(self):
        """This method will be used for the Batch Tint Settings window.
        It will define the window, as well as raise the window if it is already defined"""

        if self.batch_tint_settings_window is None:
            batchtint_filename = os.path.join(self.SETTINGS_DIR, 'batchtint_settings_filename.json')
            self.batch_tint_settings_window = Settings_Window(settings_fname=batchtint_filename)
            self.batch_tint_settings_window.backbtn.clicked.connect(lambda: raise_w(self,
                                                                                    self.batch_tint_settings_window))
            self.batch_tint_settings_window.backbtn2.clicked.connect(lambda: raise_w(self,
                                                                                     self.batch_tint_settings_window))

    def open_batch_tint_settings(self):

        if self.batch_tint_settings_window is None:
            self.set_batch_tint_settings()

        self.batch_tint_settings_window.raise_window()

    def changed_directory(self):

        self.directory_changed = True
        self.change_directory_time = time.time()
        # Find the sessions, and populate the conversion queue

    def AppendLog(self, message):
        """
        A function that will append the Log field of the main window (mainly
        used as a slot for a custom pyqt signal)
        """
        if '#' in message:
            message = message.split('#')
            color = message[-1].lower()
            message = message[0]
            message = '<span style="color:%s">%s</span>' % (color, message)

        self.log.append(message)

    def raiseError(self, error_val):
        '''raises an error window given certain errors from an emitted signal'''

        if 'NoDir' in error_val:
            self.choice = QtWidgets.QMessageBox.question(self, "No Chosen Directory",
                                                     "You have not chosen a directory,\n"
                                                     "please choose one to continue!",
                                                     QtWidgets.QMessageBox.Ok)

        elif 'NoPos' in error_val:
            session_pos_filename = error_val[error_val.find('!')+1:]
            self.choice = QtWidgets.QMessageBox.question(self, "No '.pos' file!",
                                                     "There '.pos' file for this '.rhd' session:\n" +
                                                     '%s\n' %session_pos_filename +
                                                     "was not found. would you like a dummy '.pos' file \n"
                                                     "to be created for you?\n",
                                                     QtWidgets.QMessageBox.Yes, QtWidgets.QMessageBox.No)

        elif 'NoTintSettings' in error_val:
            self.choice = QtWidgets.QMessageBox.question(self, "No Batch-Tint Settings Directory!",
                                                     "You have not chosen the settings directory for Batch-Tint,\n"
                                                     "in the main directory that holds all the Batch-Tint files\n"
                                                     "there will be a directory entitled 'settings' that holds\n"
                                                     "all the '.json' files, please choose this folder to continue!",
                                                     QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Abort)
            while self.choice is None:
                time.sleep(0.5)

            if self.choice == QtWidgets.QMessageBox.Ok:
                self.new_settings_directory()
            else:
                return

        elif 'StillNoTintSettings' in error_val:
            self.choice = QtWidgets.QMessageBox.question(self, "No Batch-Tint Settings Directory!",
                                                     "You still have not chosen the settings directory for Batch-Tint,\n"
                                                     "please choose it now.\n",
                                                     QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Abort)
            while self.choice is None:
                time.sleep(0.5)

            if self.choice == QtWidgets.QMessageBox.Ok:
                self.new_settings_directory()
            else:
                return

        elif 'DefaultTintSettings' in error_val:
            self.choice = QtWidgets.QMessageBox.question(self, "No Batch-Tint Settings Directory!",
                                                     "You still have not chosen the settings directory for Batch-Tint,\n"
                                                     "Do you want to try with the default batch-tint settings?\n",
                                                     QtWidgets.QMessageBox.Yes, QtWidgets.QMessageBox.Abort)

        elif 'InvalidTintSettings' in error_val:
            self.choice = QtWidgets.QMessageBox.question(self, "Invalid Batch-Tint Settings file!",
                                                     "You chose an invalid Batch-Tint settings directory,\n"
                                                     "Do you want to choose another directory?\n"
                                                     "Note: Press Default to use the default Batch-Tint Settings!",
                                                     QtWidgets.QMessageBox.Yes, QtWidgets.QMessageBox.Default | QtWidgets.QMessageBox.Abort)

            if self.choice == QtWidgets.QMessageBox.Yes:
                self.new_settings_directory()
            else:
                return

    def close_app(self):

        # pop up window that asks if you really want to exit the app ------------------------------------------------

        choice = QtWidgets.QMessageBox.question(self, "Quitting ",
                                            "Do you really want to exit?",
                                            QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
        if choice == QtWidgets.QMessageBox.Yes:
            sys.exit()  # tells the app to quit
        else:
            pass

    def Convert(self):
        self.choice = None
        # self.current_session = None
        self.current_subdirectory = None

        self.convert_button.setText('Stop Conversion')
        self.convert_button.setToolTip('Click to stop the conversion.')  # defining the tool tip for the start button
        self.convert_button.clicked.disconnect()
        self.convert_button.clicked.connect(self.StopConversion)

        self.conversion = True
        self.position_overwritten = False
        # start conversion threads

        self.convert_thread.start()

        self.convert_thread_worker = Worker(self.convert_queue)
        self.convert_thread_worker.moveToThread(self.convert_thread)
        self.convert_thread_worker.start.emit("start")

    def convert_queue(self):

        if default_filename in self.directory_edit.text():
            self.LogError.myGUI_signal.emit('NoDir')
            self.StopConversion()
            return

        if self.recording_queue.topLevelItemCount() == 0:
            pass

        while self.conversion:

            self.session_item = self.recording_queue.topLevelItem(0)

            if not self.session_item:
                continue
            else:
                # check if the path exists
                sessionpath = os.path.join(self.directory_edit.text(), self.session_item.data(0, 0))
                if not os.path.exists(sessionpath):
                    self.top_level_taken = False
                    self.RemoveQueueItem.myGUI_signal.emit(str(0))
                    while not self.top_level_taken:
                        time.sleep(0.1)
                    continue

            self.current_subdirectory = sessionpath

            # overwrite the settings file so it saves these settings for next time
            self.settings = self.getSettings()
            with open(self.settings_fname, 'w') as f:
                json.dump(self.settings, f)

            # overwrite the directory name
            with open(self.directory_settings, 'w') as f:
                json.dump({'directory': self.directory_edit.text()}, f)

            ConvertSession(self.session_item.data(0, 0), self.settings, self)

    def StopConversion(self):
        self.convert_button.setText('Convert')
        self.convert_button.setToolTip('Click to start the conversion.')  # defining the tool tip for the start button
        self.convert_button.clicked.disconnect()
        self.convert_button.clicked.connect(self.Convert)

        # self.convert_thread.quit()
        self.convert_thread.terminate()
        self.LogAppend.myGUI_signal.emit(
            '[%s %s]: Conversion terminated!' %
            (str(datetime.datetime.now().date()),
             str(datetime.datetime.now().time())[:8]))

        self.conversion = False

    def new_directory(self):
        """
        A function that will be used from the Choose Set popup window that will
        produce a popup so the user can pick a filename for the .set file
        """
        # prompt user to pick a .set file

        current_directory_name = QtWidgets.QFileDialog.getExistingDirectory(self, "Select a Directory!")

        # if no file chosen, skip
        if current_directory_name == '':
            return

        # change the line-edit that contains the directory information
        self.directory_edit.setText(current_directory_name)

    def new_file(self):
        """
        this method is no longer necessary, decided to have
        the user choose the settings directory instead of the file
        """

        cur_file_name, file_ext = QtWidgets.QFileDialog.getOpenFileName(self, "Select your Batch-Tint Settings File!", '',
                                              'Settings Files (*settings.json)')

        # if no file chosen, skip
        if cur_file_name == '':
            return

        # replace the current .set field in the choose .set window with chosen filename
        self.batchtintsettings_edit.setText(cur_file_name)
        self.file_chosen = True

    def new_settings_directory(self):
        current_directory_name = QtWidgets.QFileDialog.getExistingDirectory(self, "Select a Directory!")

        if current_directory_name != '':
            # replace the current .set field in the choose .set window with chosen filename
            self.batchtintsettings_edit.setText(current_directory_name)
            self.SETTINGS_DIR = current_directory_name

        self.directory_chosen = True

    def takeTopLevel(self, item_count):
        item_count = int(item_count)
        self.recording_queue.takeTopLevelItem(item_count)
        self.top_level_taken = True

    def setChild(self, child_count):
        """
        this method will set the self.child_session value to a provided child number of the
        self.session_item item.

        :param child_count:
        :return:
        """
        self.child_session = self.session_item.child(int(child_count)).clone()
        self.child_set = True

    def takeChild(self, child_count):
        self.child_session = self.session_item.takeChild(int(child_count)).clone()
        self.child_taken = True
        # return child_session

    def takeChildData(self, child_count):
        self.child_session = self.session_item.takeChild(int(child_count)).data(0, 0)
        self.child_data_taken = True

    def removeChild(self, QTreeWidgetItem):
        root = self.recording_queue.invisibleRootItem()
        (QTreeWidgetItem.parent() or root).removeChild(QTreeWidgetItem)
        self.child_removed = True


def ConvertSession(directory, bin_settings, self=None):
    """
    This function will be run on all the QTreeWidgetItems which contain the directory and session names for files that
    we want converted.

    :param directory: string, containing the name of the directory that is currently being analyzed
    :param bin_settings: contains settings, such as if the user wants to batchtint their data, or not.
    :param self: object, reference to the class's self.
    :return:
    """
    if default_filename in directory:
        if self is None:
            self.LogError.myGUI_signal.emit('NoDir')

    """This function will take in a session files and then convert the files associated with this session"""

    ##################################################################################################
    # ------------------Concatenates all the LFP/time data for each session---------------------------
    ##################################################################################################

    # main_window.current_session = directory

    session_aborted = False

    # remove the appropriate session from the TreeWidget
    iterator = QtWidgets.QTreeWidgetItemIterator(self.recording_queue)
    item_found = False
    # loops through the tree to see if the session is already there

    while iterator.value() and not item_found:

        # find the next file in the list via the iterator, in this case this will represent the directory
        self.item = iterator.value()

        # find all sessions within the directory that we will need to analyze

        if self.item.data(0, 0) == directory:
            # iterate through each item in the queue to find the appropriate directory to analyze. We will also find
            # the item_count which will help us later remove this directory once we have finalized the conversion.
            for item_count in range(self.recording_queue.topLevelItemCount()):
                if self.item == self.recording_queue.topLevelItem(item_count):
                    # these items match, therefore
                    item_found = True  # this will insure that we break the parented while loop early.

                    # initialize the tint_basenames list which will contain session basenames to analyze
                    tint_basenames = []
                    # adding the .set files to a list of session_files
                    for child_count in range(self.item.childCount()):
                        tint_basenames.append(self.item.child(child_count).data(0, 0))
                    break
        else:
            iterator += 1

    # iterate through each of the basenames within the directory
    for child_index, basename in enumerate(tint_basenames):

        if self.item.child(0).data(0, 0) == basename:
            self.child_set = False  # setting the child_set value to false

            # set the self.child_session value
            self.SetSessionItem.myGUI_signal.emit(str(0))

            # the self.child_set value will be set to True once the child has been set
            while not self.child_set:
                time.sleep(0.1)

        else:
            # considering the order through which these tint_basename values are set
            # we should not have to worry about the above if statement being false.
            # theoretically we should be performing a for loop of all the children
            # within this item and finding the item that matches.
            pass

        # obtaining the corresponding .set filename
        set_filename = os.path.join(self.directory_edit.text(), directory, basename)

        # convert the corresponding .bin file of the obtained .set file, if any errors have occurred the returned
        # converted value should be an 'Aborted' string value.
        converted = convert_basename(self, set_filename, bin_settings['threshold'])

        # remove the session's QTreeWidgetItem since we have analyzed it already
        self.child_taken = False  # initialize the child taken value
        self.RemoveSessionItem.myGUI_signal.emit(str(0))  # remove the child

        # wait for child to be removed
        while not self.child_taken:
            time.sleep(0.1)

        # set the child_session to None since we are done converting the previous session
        self.child_session = None

        # check if there are any child items within the directory, if not then we can remove the directory from
        # the queue
        if self.item.childCount() == 0:
            self.top_level_taken = False
            self.RemoveQueueItem.myGUI_signal.emit(str(item_count))
            # wait for directory item to be removed
            while not self.top_level_taken:
                time.sleep(0.1)

        # check if there were any errors in the conversion
        if isinstance(converted, str):
            # ensures that the aborted session does not go into batchTint or get moved
            if 'Aborted' in converted:
                session_aborted = True
                continue
                # return

    if session_aborted:
        # in the case that a session was aborted we simply finish the conversion process. Maybe in the future we
        # should just continue on our way and just remove this session from the queue. It is possible that this session
        # is corrupted.
        return

    # determine if the user wants to sort the newly converted data
    if bin_settings['batchtint']:
        # if batch-tint is checked, don't move to converted, just convert run BatchTINTV3 here

        # import the settings values
        with open(self.batch_tint_settings_window.settings_fname, 'r') as f:
            settings = json.load(f)

        settings['move_processed'] = 0  # we will override the move_processed functionality here.

        smtp_settings = {}
        smtp_settings['Notification'] = 0  # 1 for send e-mails, 0 for don't send

        '''
        # if you have the notifications set to 0 you don't have to worry about this.
        # we will need an e-mail to send these experimenter's e-mails from
        smtp_settings['Username'] = 'example@gmail.com'
        smtp_settings['Password'] = 'password'  # associated password
        smtp_settings['ServerName'] = 'smtp.gmail.com'  # the smtp server name, 'smtp.gmail.com' for gmail
        smtp_settings['Port'] = 587  # 587 default for gmail
        '''

        experimenter_settings = {
            # 'example': 'example@gmail.com'  # can do [email1@.., email2@..] if you want it sent to more than 1
        }

        analyzed_set_files = klusta([set_filename], settings,
                                    smtp_settings=smtp_settings,
                                    experimenter_settings=experimenter_settings,
                                    append=None, self=self)

    # move the files to a Converted folder if the user wants to
    if bin_settings['move_converted']:
        # if the user wants to move the converted files.

        # move to the converted file
        convert_fpath = os.path.join(self.directory_edit.text(), 'Converted')
        if not os.path.exists(convert_fpath):
            os.mkdir(convert_fpath)

        directory_source = os.path.join(self.directory_edit.text(), directory)
        directory_destination = os.path.join(convert_fpath, directory)

        if os.path.exists(directory_destination):
            try:
                copy_tree(directory_source, directory_destination)
            except FileNotFoundError:
                return
            try:
                shutil.rmtree(directory_source)
            except PermissionError:
                self.LogAppend.myGUI_signal.emit(
                    '[%s %s]: The current directory could not be deleted, close files and then delete the directory!#red' %
                    (str(datetime.datetime.now().date()),
                     str(datetime.datetime.now().time())[:8]))
        else:
            shutil.move(directory_source, convert_fpath)

    # finished converting
    self.LogAppend.myGUI_signal.emit(
        '[%s %s]: The following directory has finished converting: %s!' %
        (str(datetime.datetime.now().date()),
         str(datetime.datetime.now().time())[:8], self.directory_edit.text()))


def run():
    app = QtWidgets.QApplication(sys.argv)

    main_window = Window()  # calling the main window
    main_window.raise_()  # making the main window on top

    sys.exit(app.exec_())  # prevents the window from immediately exiting out


if __name__ == '__main__':
    run()




import os
import numpy as np
import struct
from .Tint_Matlab import int16toint8, get_setfile_parameter
import scipy
import scipy.signal
from .conversion_utils import get_set_header


def fir_hann(data, Fs, cutoff, n_taps=101, showresponse=0):
    # The Nyquist rate of the signal.
    nyq_rate = Fs / 2

    b = scipy.signal.firwin(n_taps, cutoff / nyq_rate, window='hann')

    a = 1.0
    # Use lfilter to filter x with the FIR filter.
    data = scipy.signal.lfilter(b, a, data)

    '''
    if showresponse == 1:
        w, h = scipy.signal.freqz(b, a, worN=8000)  # returns the requency response h, and the angular frequencies
        # w in radians/sec
        # w (radians/sec) * (1 cycle/2pi*radians) = Hz
        # f = w / (2 * np.pi)  # Hz

        plt.figure(figsize=(20, 15))
        plt.subplot(211)
        plt.semilogx((w / np.pi) * nyq_rate, np.abs(h), 'b')
        plt.xscale('log')
        plt.title('%s Filter Frequency Response')
        plt.xlabel('Frequency(Hz)')
        plt.ylabel('Gain [V/V]')
        plt.margins(0, 0.1)
        plt.grid(which='both', axis='both')
        plt.axvline(cutoff, color='green')
    '''

    return data, n_taps


def create_eeg(filename, data, Fs, DC_Blocker=True):
    # data is given in int16

    if os.path.exists(filename):
        return

    Fs_EGF = int(4.8e3)  # sampling rate of .EGF files
    Fs_EEG = int(250)

    data = data[:, 0::int(Fs / Fs_EGF)]

    # append zeros to make the duration a round number
    duration_round = np.ceil(data.shape[1] / Fs_EGF)  # the duration should be rounded up to the nearest integer
    missing_samples = int(duration_round * Fs_EGF - data.shape[1])
    if missing_samples != 0:
        missing_samples = np.tile(np.array([0]), (1, missing_samples))
        data = np.hstack((data, missing_samples))

    # ensure the full last second of data is equal to zero
    data[0, -int(Fs_EGF):] = 0

    data = data[0, :-Fs_EGF]

    # data = np.rint(data)  # convert the data to integers
    data = data.astype(np.int32)  # convert the data to integers

    # ensuring the appropriate range of the values
    data[np.where(data > 32767)] = 32767
    data[np.where(data < -32768)] = -32768

    data, N = fir_hann(data, Fs_EGF, 125, n_taps=101, showresponse=0)  # FIR filter to remove anti aliasing

    data = int16toint8(data)

    data = EEG_downsample(data)

    # append 1 second of 0's like tint does
    data = np.hstack((data.flatten(), np.zeros((250, 1)).flatten()))
    ##################################################################################################
    # ---------------------------Writing the EEG Data-------------------------------------------
    ##################################################################################################

    write_eeg(filename, data, Fs_EEG)


def EEG_downsample(EEG):
    """The EEG data is created from the EGF files which involves a 4.8k to 250 Hz conversion"""
    EEG = EEG.flatten()

    i = -1
    # i = 0

    # indices = [i]
    indices = []
    while i < len(EEG) - 1:
        indices.extend([(i + 19), (i + 19 * 2), (i + 19 * 3), (i + 19 * 4), (i + 19 * 4 + 20)])
        # indices.extend([(i+20), (i+20+19), (i+20+19*2), (i+20+19*3), (i+20+19*4)])
        i += (19 * 4 + 20)

    indices = np.asarray(indices)

    indices = indices[np.where(indices <= len(EEG) - 1)]

    return EEG[indices]


def create_egf(filename, data, Fs, DC_Blocker=True):
    if os.path.exists(filename):
        return

    Fs_EGF = int(4.8e3)  # sampling rate of .EGF files

    data = data[:, 0::int(Fs / Fs_EGF)]

    # notch filter the data
    # data = sp.Filtering().notch_filt(data, Fs_EGF, freq=60, band=10, order=3)

    # append zeros to make the duration a round number
    duration_round = np.ceil(data.shape[1] / Fs_EGF)  # the duration should be rounded up to the nearest integer
    missing_samples = int(duration_round * Fs_EGF - data.shape[1])
    if missing_samples != 0:
        missing_samples = np.tile(np.array([0]), (1, missing_samples))
        data = np.hstack((data, missing_samples))

    # ensure the full last second of data is equal to zero
    data[0, -int(Fs_EGF):] = 0

    # data = np.rint(data)  # convert the data to integers
    data = data.astype(np.int32)  # convert the data to integers

    # ensuring the appropriate range of the values
    data[np.where(data > 32767)] = 32767
    data[np.where(data < -32768)] = -32768

    # data is already in int16 which is what the final unit should be in

    ##################################################################################################
    # ---------------------------Writing the EGF Data-------------------------------------------
    ##################################################################################################

    write_eeg(filename, data, Fs_EGF)


def write_eeg(filepath, data, Fs):

    data = data.flatten()

    session_path, session_filename = os.path.split(filepath)

    tint_basename = os.path.splitext(session_filename)[0]

    set_filename = os.path.join(session_path, '%s.set' % tint_basename)
    header = get_set_header(set_filename)

    num_samples = int(len(data))

    if '.egf' in session_filename:

        # EEG_Fs = 4800
        egf = True

    else:

        # EEG_Fs = 250
        egf = False

    # if the duration before the set file was overwritten wasn't a round number, it rounded up and thus we need
    # to append values to the EEG (we will add 0's to the end)
    duration = int(get_setfile_parameter('duration', set_filename))  # get the duration from the set file

    EEG_expected_num = int(Fs * duration)

    if num_samples < EEG_expected_num:
        missing_samples = EEG_expected_num - num_samples
        data = np.hstack((data, np.zeros((1, missing_samples)).flatten()))
        num_samples = EEG_expected_num

    with open(filepath, 'w') as f:

        num_chans = 'num_chans 1'

        if egf:
            sample_rate = '\nsample_rate %d Hz' % (int(Fs))
            data = struct.pack('<%dh' % (num_samples), *[np.int(data_value) for data_value in data.tolist()])
            b_p_sample = '\nbytes_per_sample 2'
            num_EEG_samples = '\nnum_EGF_samples %d' % (num_samples)

        else:
            sample_rate = '\nsample_rate %d.0 hz' % (int(Fs))
            data = struct.pack('>%db' % (num_samples), *[np.int(data_value) for data_value in data.tolist()])
            b_p_sample = '\nbytes_per_sample 1'
            num_EEG_samples = '\nnum_EEG_samples %d' % (num_samples)

        eeg_p_position = '\nEEG_samples_per_position %d' % (5)

        start = '\ndata_start'

        if egf:
            write_order = [header, num_chans, sample_rate,
                           b_p_sample, num_EEG_samples, start]
        else:
            write_order = [header, num_chans, sample_rate, eeg_p_position,
                           b_p_sample, num_EEG_samples, start]

        f.writelines(write_order)

    with open(filepath, 'rb+') as f:
        f.seek(0, 2)
        f.writelines([data, bytes('\r\ndata_end\r\n', 'utf-8')])




import os
from .conversion_utils import get_set_header
from .Tint_Matlab import get_setfile_parameter
import struct


def create_pos(pos_filename, pos_data):
    n = int(pos_data.shape[0])

    session_path, session_filename = os.path.split(pos_filename)
    tint_basename = os.path.splitext(session_filename)[0]
    set_filename = os.path.join(session_path, '%s.set' % tint_basename)

    if os.path.exists(pos_filename):
        return

    header = get_set_header(set_filename)

    with open(pos_filename, 'wb+') as f:  # opening the .pos file
        write_list = []

        [write_list.append(bytes('%s\r\n' % value, 'utf-8')) for value in header.split('\n') if value != '']

        header_vals = ['num_colours %d' % 4]
        if 'abid' in header.lower():
            header_vals.extend(
                ['\r\nmin_x %d' % 0,
                 '\r\nmax_x %d' % 768,
                 '\r\nmin_y %d' % 0,
                 '\r\nmax_y %d' % 574])
        elif 'gus' in header.lower():
            header_vals.extend(
                ['\r\nmin_x %d' % 0,
                 '\r\nmax_x %d' % 640,
                 '\r\nmin_y %d' % 0,
                 '\r\nmax_y %d' % 480]

            )
        header_vals.extend(
            ['\r\nwindow_min_x %d' % int(get_setfile_parameter('xmin', set_filename)),
            '\r\nwindow_max_x %d' % int(get_setfile_parameter('xmax', set_filename)),
            '\r\nwindow_min_y %d' % int(get_setfile_parameter('ymin', set_filename)),
            '\r\nwindow_max_y %d' % int(get_setfile_parameter('ymax', set_filename)),
            '\r\ntimebase %d hz' % 50,
            '\r\nbytes_per_timestamp %d' % 4,
            '\r\nsample_rate %.1f hz' % 50.0,
            '\r\nEEG_samples_per_position %d' % 5,
            '\r\nbearing_colour_1 %d' % 0,
            '\r\nbearing_colour_2 %d' % 0,
            '\r\nbearing_colour_3 %d' % 0,
            '\r\nbearing_colour_4 %d' % 0,
            '\r\npos_format t,x1,y1,x2,y2,numpix1,numpix2',
            '\r\nbytes_per_coord %d' % 2,
            '\r\npixels_per_metre %f' % float(
               get_setfile_parameter('tracker_pixels_per_metre', set_filename)),
            '\r\nnum_pos_samples %d' % n,
            '\r\ndata_start'])

        for value in header_vals:
            write_list.append(bytes(value, 'utf-8'))

        onespot = 1  # this is just in case we decide to add other modes.

        # write_list = [bytes(headers, 'utf-8')]

        # write_list.append()

        if onespot:
            position_format_string = 'i8h'
            position_format_string = '>%s' % (n * position_format_string)
            write_list.append(struct.pack(position_format_string, *pos_data.astype(int).flatten()))

        write_list.append(bytes('\r\ndata_end\r\n', 'utf-8'))
        f.writelines(write_list)



from __future__ import division, print_function
import numpy as np
import struct, os
import numpy.matlib
from scipy.io import savemat
import mmap
import contextlib


def int16toint8(value):
    """Converts int16 data to int8"""
    value = np.divide(value, 256).astype(int)
    value[np.where(value > 127)] = 127
    value[np.where(value < -128)] = -128

    return value


def remEEGShift(t, EEG):
    """This function removes the shift that the EEG undergoes"""
    EEG = EEG.flatten()
    Fs_EGF = int(4.8e3)
    # normal delay is 0.5(n_taps-1) / Fs_EGF, n_taps is 101 for the FIR filter
    # there is already a shift of 18 as the EEG starts off using the 18th EGF data
    delay = (0.5 * (101 - 1) - 18) / Fs_EGF

    t = t - delay

    t_bool = np.where(t >= 0)[0]
    t = t[t_bool]
    EEG = EEG[t_bool]

    return t, EEG


def MatlabNumSeq(start, stop, step, exclude=True):
    """In Matlab you can type:

    start:step:stop and easily create a numerical sequence

    if exclude is true it will exclude any values greater than the stop value
    """

    '''np.arange(start, stop, step) works good most of the time

    However, if the step (stop-start)/step is an integer, then the sequence
    will stop early'''

    seq = np.arange(start, stop + step, step)

    if exclude:
        if seq[-1] > stop:
            seq = seq[:-1]

    return seq


class TintException(Exception):
    def __init___(self, message):
        Exception.__init__(self, "%s" % message)
        self.message = message


def get_good_cells(units):
    """
    In Tint we have a method of determining the 'good cells'. A good tetrode must begin
    with cell #1, and the set of consecutive cell numberes are all good. Essentially we leave
    a blank cell separating the good cells from the bad. This function will return the good cells.
    """
    units = np.unique(units)

    if 1 not in units:
        return np.array([])

    good_cells = [1]

    found_separation = False
    for index, unit in enumerate(units):
        if unit > 1:
            if unit == units[index - 1] + 1 and not found_separation:
                good_cells.append(unit)
            else:  #
                # no longer consecutive units, so set value to false to ensure no more appending
                found_separation = True

    return np.asarray(good_cells)


def get_setfile_parameter(parameter, set_filename):
    """
    This function will return the parameter value of a given parameter name for a given set filename.

    Example:
        set_fullpath = 'C:\\example\\tetrode_1.1'
        parameter_name = 'duration
        duration = get_setfile_parameter(parameter_name, set_fullpath)

    Args:
        parameter (str): the name of the set file parameter that you want to obtain.
        set_filename (str): the full path of the .set file that you want to obtain the parameter value from.


    Returns:
        parameter_value (str): the value for the given parameter

    """

    if not os.path.exists(set_filename):
        return

    # adding the encoding because tint data is created via windows and if you want to run this in linux, you need
    # to explicitly say this
    with open(set_filename, 'r+', encoding='cp1252') as f:
        for line in f:
            if parameter in line:
                if line.split(' ')[0] == parameter:
                    # prevents part of the parameter being in another parameter name
                    new_line = line.strip().split(' ')
                    if len(new_line) == 2:
                        return new_line[-1]
                    else:
                        return ' '.join(new_line[1:])


def getpos(pos_fpath, arena, method='', flip_y=True):
    '''getpos function:
    ---------------------------------------------
    variables:
    -pos_fpath: the full path (C:\example\session.pos)

    output:
    t: column numpy array of the time stamps
    x: a column array of the x-values (in pixels)
    y: a column array of the y-values (in pixels)
    '''

    with open(pos_fpath, 'rb+') as f:  # opening the .pos file
        headers = ''  # initializing the header string
        for line in f:  # reads line by line to read the header of the file
            # print(line)
            if 'data_start' in str(line):  # if it reads data_start that means the header has ended
                headers += 'data_start'
                break  # break out of for loop once header has finished
            elif 'duration' in str(line):
                headers += line.decode(encoding='UTF-8')
            elif 'num_pos_samples' in str(line):
                num_pos_samples = int(line.decode(encoding='UTF-8')[len('num_pos_samples '):])
                headers += line.decode(encoding='UTF-8')
            elif 'bytes_per_timestamp' in str(line):
                bytes_per_timestamp = int(line.decode(encoding='UTF-8')[len('bytes_per_timestamp '):])
                headers += line.decode(encoding='UTF-8')
            elif 'bytes_per_coord' in str(line):
                bytes_per_coord = int(line.decode(encoding='UTF-8')[len('bytes_per_coord '):])
                headers += line.decode(encoding='UTF-8')
            elif 'timebase' in str(line):
                timebase = (line.decode(encoding='UTF-8')[len('timebase '):]).split(' ')[0]
                headers += line.decode(encoding='UTF-8')
            elif 'pixels_per_metre' in str(line):
                ppm = float(line.decode(encoding='UTF-8')[len('pixels_per_metre '):])
                headers += line.decode(encoding='UTF-8')
            elif 'min_x' in str(line) and 'window' not in str(line):
                min_x = int(line.decode(encoding='UTF-8')[len('min_x '):])
                headers += line.decode(encoding='UTF-8')
            elif 'max_x' in str(line) and 'window' not in str(line):
                max_x = int(line.decode(encoding='UTF-8')[len('max_x '):])
                headers += line.decode(encoding='UTF-8')
            elif 'min_y' in str(line) and 'window' not in str(line):
                min_y = int(line.decode(encoding='UTF-8')[len('min_y '):])
                headers += line.decode(encoding='UTF-8')
            elif 'max_y' in str(line) and 'window' not in str(line):
                max_y = int(line.decode(encoding='UTF-8')[len('max_y '):])
                headers += line.decode(encoding='UTF-8')
            elif 'pos_format' in str(line):
                headers += line.decode(encoding='UTF-8')
                if 't,x1,y1,x2,y2,numpix1,numpix2' in str(line):
                    two_spot = True
                else:
                    two_spot = False
                    print('The position format is unrecognized!')

            elif 'sample_rate' in str(line):
                sample_rate = float(line.decode(encoding='UTF-8').split(' ')[1])
                headers += line.decode(encoding='UTF-8')

            else:
                headers += line.decode(encoding='UTF-8')

    if two_spot:

        '''Run when two spot mode is on, (one_spot has the same format so it will also run here)'''
        with open(pos_fpath, 'rb+') as f:
            '''get_pos for one_spot'''
            pos_data = f.read()  # all the position data values (including header)
            pos_data = pos_data[len(headers):-12]  # removes the header values

            byte_string = 'i8h'

            pos_data = np.asarray(struct.unpack('>%s' % (num_pos_samples * byte_string), pos_data))
            pos_data = pos_data.astype(float).reshape((num_pos_samples, 9))  # there are 8 words and 1 time sample

        x = pos_data[:, 1]
        y = pos_data[:, 2]
        t = pos_data[:, 0]

        x = x.reshape((len(x), 1))
        y = y.reshape((len(y), 1))
        t = t.reshape((len(t), 1))

        if method == 'raw':
            return x, y, t, sample_rate

        t = np.divide(t, np.float(timebase))  # converting the frame number from Axona to the time value

        # values that are NaN are set to 1023 in Axona's system, replace these values by NaN's

        x[np.where(x == 1023)] = np.nan
        y[np.where(y == 1023)] = np.nan

        didFix, fixedPost = fixTimestamps(t)

        if didFix:
            t = fixedPost

        t = t - t[0]

        x, y = arena_config(x, y, arena, conversion=ppm, center=np.asarray([np.mean([np.amin(x), np.amax(x)]),
                                                                            np.mean([np.amin(y), np.amax(y)])]),
                            flip_y=flip_y)

        # remove any NaNs at the end of the file
        x, y, t = removeNan(x, y, t)

    else:
        print("Haven't made any code for this part yet.")

    return x.reshape((len(x), 1)), y.reshape((len(y), 1)), t.reshape((len(t), 1)), sample_rate


def is_tetrode(file, session):
    """"
    Determines if the file is a tetrode, essentially will look at the extension and see if it ends in an integer
    or not. Also checks that the tetrode belongs to a given session.
    """
    if os.path.splitext(file)[0] == session:
        try:
            tetrode_number = int(os.path.splitext(file)[1][1:])
            return True
        except ValueError:
            return False
    else:
        return False


def get_active_tetrode(set_filename):
    """in the .set files it will say collectMask_X Y for each tetrode number to tell you if
    it is active or not. T1 = ch1-ch4, T2 = ch5-ch8, etc."""
    active_tetrode = []
    active_tetrode_str = 'collectMask_'

    with open(set_filename) as f:
        for line in f:

            # collectMask_X Y, where x is the tetrode number, and Y is eitehr on or off (1 or 0)
            if active_tetrode_str in line:
                tetrode_str, tetrode_status = line.split(' ')
                if int(tetrode_status) == 1:
                    # then the tetrode is saved
                    tetrode_str.find('_')
                    tet_number = int(tetrode_str[tetrode_str.find('_') + 1:])
                    active_tetrode.append(tet_number)

    return active_tetrode


def get_active_eeg(set_filename):
    """This will return a dictionary (cative_eeg_dict) where the keys
    will be eeg channels from 1->64 which will represent the eeg suffixes (2 = .eeg2, 3 = 2.eeg3, etc)
    and the key will be the channel that the EEG maps to (a channel from 0->63)"""
    active_eeg = []
    active_eeg_str = 'saveEEG_ch'

    eeg_map = []
    eeg_map_str = 'EEG_ch_'

    active_eeg_dict = {}

    with open(set_filename) as f:
        for line in f:

            if active_eeg_str in line:
                # saveEEG_ch_X Y, where x is the eeg number, and Y is eitehr on or off (1 or 0)
                _, status = line.split(' ')
                active_eeg.append(int(status))
            elif eeg_map_str in line:
                # EEG_ch_X Y
                _, chan = line.split(' ')
                eeg_map.append(int(chan))

                # active_eeg = np.asarray(active_eeg)
                # eeg_map = np.asarray(eeg_map)

    for i, status in enumerate(active_eeg):
        if status == 1:
            active_eeg_dict[i + 1] = eeg_map[i] - 1

    return active_eeg_dict


def is_egf_active(set_filename):
    active_egf_str = 'saveEGF'

    with open(set_filename) as f:
        for line in f:

            if active_egf_str in line:
                _, egf_status = line.split(' ')

                if int(egf_status) == 1:
                    return True

        return False


def find_tetrodes(set_fullpath):
    """finds the tetrode files available for a given .set file if there is a  .cut file existing"""

    tetrode_path, session = os.path.split(set_fullpath)
    session, _ = os.path.splitext(session)

    # getting all the files in that directory
    file_list = os.listdir(tetrode_path)

    # acquiring only a list of tetrodes that belong to that set file
    tetrode_list = [os.path.join(tetrode_path, file) for file in file_list
                    if is_tetrode(file, session)]

    # if the .cut file doesn't exist remove list

    tetrode_list = [file for file in tetrode_list if os.path.exists(
        os.path.join(tetrode_path, '%s_%s.cut' % (os.path.splitext(file)[0], os.path.splitext(file)[1][1:])))]

    return tetrode_list


def find_unit(tetrode_list):
    """Inputs:
    tetrode_list: list of tetrodes to find the units that are in the tetrode_path
    example [r'C:Location\of\File\filename.1', r'C:Location\of\File\filename.2' ],
    -------------------------------------------------------------
    Outputs:
    cut_list: an nx1 list for n-tetrodes in the tetrode_list containing a list of unit numbers that each spike belongs to
    """

    input_list = True
    if type(tetrode_list) != list:
        input_list = False
        tetrode_list = [tetrode_list]

    cut_list = []
    unique_cell_list = []
    for tetrode_file in tetrode_list:
        directory = os.path.dirname(tetrode_file)

        try:
            tetrode = int(os.path.splitext(tetrode_file)[1][1:])
        except ValueError:
            raise ValueError("The following file is invalid: %s" % tetrode_file)

        tetrode_base = os.path.splitext(os.path.basename(tetrode_file))[0]

        cut_filename = os.path.join(directory, '%s_%d.cut' % (tetrode_base, tetrode))

        cut_values = read_cut(cut_filename)
        cut_list.append(cut_values)
        unique_cell_list.append(np.unique(cut_values))

    return cut_list


def read_clu(filename):
    """
    This will read in the .clu.N files that are provided by Tint. The .clu cell ID's go from 1 -> N
    instead of the traditional 0->N-1 for a .cut file. We will convert from the 1->N format to the
    0->N-1 format.
    """

    data = np.loadtxt(filename)

    # the first number in the file is simply the number of cells that were recorded, we must remove this

    # we will also subtract 1 to ensure that the data goes from 0->N-1 instead of 1->N. Essentially converting
    # from the clu format to the .cut format.

    return data[1:].flatten() - 1


def read_cut(cut_filename):
    """This function will read the given cut file, and output the """
    cut_values = None
    if os.path.exists(cut_filename):
        extract_cut = False
        with open(cut_filename, 'r') as f:
            for line in f:
                if 'Exact_cut' in line:  # finding the beginning of the cut values
                    extract_cut = True
                if extract_cut:  # read all the cut values
                    cut_values = str(f.readlines())
                    for string_val in ['\\n', ',', "'", '[', ']']:  # removing non base10 integer values
                        cut_values = cut_values.replace(string_val, '')
                    cut_values = [int(val) for val in cut_values.split()]
        cut_values = np.asarray(cut_values)
    return cut_values


def arena_config(posx, posy, arena, conversion='', center='', flip_y=True):
    """
    :param posx:
    :param posy:
    :param arena:
    :param conversion:
    :param center:
    :param flip_y: bool value that will determine if you want to flip y or not. When recording on Intan we inverted the
    positions due to the camera position. However in the virtualmaze you might not want to flip y values.
    :return:
    """
    if 'BehaviorRoom' in arena:
        center = np.array([314.75, 390.5])
        conversion = 495.5234
    elif 'BehaviorRoom2' in arena:
        # added december 12th, 2018
        center = np.array([314.75, 390.5])
        conversion = 485.1185
    elif 'DarkRoom' in arena:
        center = np.array([346.5, 273.5])
        conversion = 711.3701
    elif 'room4' in arena:
        center = np.array([418, 186])
        conversion = 313
    elif arena in ['Linear Track', 'Circular Track', 'Four Leaf Clover Track', 'Simple Circular Track',
                   'Parallel Linear Global Track', 'Parallel Linear Rate Track']:
        center = center
        conversion = conversion
    else:
        print("Room: " + arena + ", is an unknown room!")

    posx = 100 * (posx - center[0]) / conversion

    if flip_y:
        posy = 100 * (-posy + center[1]) / conversion

    return posx, posy


def ReadEEG(eeg_fname):
    """input:
    eeg_filename: the fullpath to the eeg file that is desired to be read.
    Example: C:\Location\of\eegfile.eegX

    Output:
    The EEG waveform, and the sampling frequency"""

    with open(eeg_fname, 'rb') as f:

        is_eeg = False
        if 'eeg' in eeg_fname:
            is_eeg = True
            # Fs = 250
        # else:
        #    Fs = 4.8e3

        with contextlib.closing(mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ)) as m:
            # find the data_start
            start_index = int(m.find(b'data_start') + len('data_start'))  # start of the data
            stop_index = int(m.find(b'\r\ndata_end'))  # end of the data

            sample_rate_start = m.find(b'sample_rate')
            sample_rate_end = m[sample_rate_start:].find(b'\r\n')
            Fs = float(m[sample_rate_start:sample_rate_start + sample_rate_end].decode('utf-8').split(' ')[1])

            m = m[start_index:stop_index]

            if is_eeg:
                EEG = np.fromstring(m, dtype='>b')
            else:
                EEG = np.fromstring(m, dtype='<h')

            return EEG, int(Fs)


def EEG_to_Mat(input_filename, output_filename):
    EEG, Fs = ReadEEG(input_filename)

    if Fs > 250:
        save_dictionary = {'EGF': EEG, 'Fs': Fs}
    else:
        save_dictionary = {'EEG': EEG, 'Fs': Fs}

    savemat(output_filename, save_dictionary)


def removeNan(posx, posy, post):
    """Remove any NaNs from the end of the array"""
    removeNan = True
    while removeNan:
        if np.isnan(posx[-1]):
            posx = posx[:-1]
            posy = posy[:-1]
            post = post[:-1]
        else:
            removeNan = False
    return posx, posy, post


def centerBox(posx, posy):
    # must remove Nans first because the np.amin will return nan if there is a nan
    posx = posx[~np.isnan(posx)]  # removes NaNs
    posy = posy[~np.isnan(posy)]  # remove Nans

    NE = np.array([np.amax(posx), np.amax(posy)])
    NW = np.array([np.amin(posx), np.amax(posy)])
    SW = np.array([np.amin(posx), np.amin(posy)])
    SE = np.array([np.amax(posx), np.amin(posy)])

    return findCenter(NE, NW, SW, SE)


def findCenter(NE, NW, SW, SE):
    """Finds the center point (x,y) of the position boundaries"""

    x = np.mean([np.amax([NE[0], SE[0]]), np.amin([NW[0], SW[0]])])
    y = np.mean([np.amax([NW[1], NE[1]]), np.amin([SW[1], SE[1]])])
    return np.array([x, y])


def bits2uV(data, data_fpath, set_fpath=''):
    '''

    :param data:
    :param data_fpath: example: 'C:\example\filepath.whatever'
    :param set_fpath:
    :return:
    '''
    path = os.path.split(data_fpath)[0]

    if set_fpath == '':
        set_fpath = os.path.join(path, ''.join([os.path.splitext(os.path.basename(data_fpath))[0],'.set']))

    ext = os.path.splitext(data_fpath)[1]

    if not os.path.exists(set_fpath):
        error_message = 'The following setpath does not exist, cannot convert to uV: %s' % (set_fpath)
        raise TintException(error_message)
        #return error_message, 0

    # create a tetrode map that has rows of channels that correspond to the same tetrode
    tet_map = np.asarray([np.arange(start,start+4) for start in np.arange(0, 32)*4])

    chan_gains = np.array([])
    saved_eeg = np.array([])
    eeg_chan_map = np.array([])

    with open(set_fpath, 'r') as f:
        for line in f:

            if 'ADC_fullscale_mv' in line:
                ADC_fullscale_mv = int(line.split(" ")[1])
            elif 'gain_ch_' in line:
                # create an array of channel gains [channel_number, channels_gain]
                if len(chan_gains) == 0:
                    chan_gains = np.array([int(line[len('gain_ch_'):line.find(" ")]), int(line.split(" ")[1])], ndmin=2)
                else:
                    chan_gains = np.append(chan_gains, np.array([int(line[len('gain_ch_'):line.find(" ")]), int(line.split(" ")[1])], ndmin=2), axis=0)
            elif 'saveEEG_ch_' in line:
                # create an array of EEG channels that are saved
                if int(line.split(" ")[1]) == 1:
                    if len(chan_gains) == 0:
                        saved_eeg = np.array([int(line[len('saveEEG_ch_'):line.find(" ")])])
                    else:
                        saved_eeg = np.append(saved_eeg, np.array([int(line[len('saveEEG_ch_'):line.find(" ")])]))
            elif 'EEG_ch_' in line and 'BPF' not in line:
                if len(eeg_chan_map) == 0:
                    eeg_chan_map = np.array([int(line[len('EEG_ch_'):line.find(" ")]), int(line.split(" ")[1])], ndmin=2)
                else:
                    eeg_chan_map = np.append(eeg_chan_map, np.array([int(line[len('EEG_ch_'):line.find(" ")]), int(line.split(" ")[1])], ndmin=2), axis=0)

    if '.eeg' in ext:
        if len(ext) == len('.eeg'):
            chan_num = 1
        else:
            chan_num = int(ext[len('.eeg'):])

        for index, value in enumerate(eeg_chan_map[:]):
            if value[0] == chan_num:
                eeg_chan = value[1] - 1
                break

        for index, value in enumerate(chan_gains):
            if value[0] == eeg_chan:
                gain = value[1]

        scalar = ADC_fullscale_mv*1000/(gain*128)
        if len(data) == 0:
            data_uV = []
        else:
            data_uV = np.multiply(data, scalar)
            #print(data_uV)

    elif '.egf' in ext:
        if len(ext) == len('.egf'):
            chan_num = 1
        else:
            chan_num = int(ext[len('.egf'):])

        for index, value in enumerate(eeg_chan_map[:]):
            if value[0] == chan_num:
                eeg_chan = value[1] - 1
                break

        for index, value in enumerate(chan_gains):
            if value[0] == eeg_chan:
                gain = value[1]
                break

        scalar = ADC_fullscale_mv*1000/(gain*32768)

        if len(data) == 0:
            data_uV = []
        else:
            data_uV = np.multiply(data, scalar)

    else:
        tetrode_num = int(ext[1:])

        tet_chans = tet_map[tetrode_num-1]

        gain = np.asarray([[gains[1] for gains in chan_gains if gains[0] == chan] for chan in tet_chans])

        scalar = (ADC_fullscale_mv*1000/(gain*128).reshape((1, len(gain))))[0]

        if len(data) == 0:
            data_uV = []
        else:
            data_uV = np.multiply(data, scalar)

    return data_uV, scalar


def getspikes(fullpath):
    """
    This function will return the spike data, spike times, and spike parameters from Tint tetrode data.

    Example:
        tetrode_fullpath = 'C:\\example\\tetrode_1.1'
        ts, ch1, ch2, ch3, ch4, spikeparam = getspikes(tetrode_fullpath)

    Args:
        fullpath (str): the fullpath to the Tint tetrode file you want to acquire the spike data from.

    Returns:
        ts (ndarray): an Nx1 array for the spike times, where N is the number of spikes.
        ch1 (ndarray) an NxM matrix containing the spike data for channel 1, N is the number of spikes,
            and M is the chunk length.
        ch2 (ndarray) an NxM matrix containing the spike data for channel 2, N is the number of spikes,
            and M is the chunk length.
        ch3 (ndarray) an NxM matrix containing the spike data for channel 3, N is the number of spikes,
            and M is the chunk length.
        ch4 (ndarray) an NxM matrix containing the spike data for channel 4, N is the number of spikes,
            and M is the chunk length.
        spikeparam (dict): a dictionary containing the header values from the tetrode file.
    """
    spikes, spikeparam = importspikes(fullpath)
    ts = spikes['t']
    nspk = spikeparam['num_spikes']
    spikelen = spikeparam['samples_per_spike']

    ch1 = spikes['ch1']
    ch2 = spikes['ch2']
    ch3 = spikes['ch3']
    ch4 = spikes['ch4']

    return ts, ch1, ch2, ch3, ch4, spikeparam


def importspikes(filename):
    """Reads through the tetrode file as an input and returns two things, a dictionary containing the following:
    timestamps, ch1-ch4 waveforms, and it also returns a dictionary containing the spike parameters"""

    with open(filename, 'rb') as f:
        for line in f:
            if 'data_start' in str(line):
                spike_data = np.fromstring((line + f.read())[len('data_start'):-len('\r\ndata_end\r\n')], dtype='uint8')
                break
            elif 'num_spikes' in str(line):
                num_spikes = int(line.decode(encoding='UTF-8').split(" ")[1])
            elif 'bytes_per_timestamp' in str(line):
                bytes_per_timestamp = int(line.decode(encoding='UTF-8').split(" ")[1])
            elif 'samples_per_spike' in str(line):
                samples_per_spike = int(line.decode(encoding='UTF-8').split(" ")[1])
            elif 'bytes_per_sample' in str(line):
                bytes_per_sample = int(line.decode(encoding='UTF-8').split(" ")[1])
            elif 'timebase' in str(line):
                timebase = int(line.decode(encoding='UTF-8').split(" ")[1])
            elif 'duration' in str(line):
                duration = int(line.decode(encoding='UTF-8').split(" ")[1])
            elif 'sample_rate' in str(line):
                samp_rate = int(line.decode(encoding='UTF-8').split(" ")[1])

                # calculating the big-endian and little endian matrices so we can convert from bytes -> decimal
    big_endian_vector = 256 ** np.arange(bytes_per_timestamp - 1, -1, -1)
    little_endian_matrix = np.arange(0, bytes_per_sample).reshape(bytes_per_sample, 1)
    little_endian_matrix = 256 ** numpy.matlib.repmat(little_endian_matrix, 1, samples_per_spike)

    number_channels = 4

    # calculating the timestamps
    t_start_indices = np.linspace(0, num_spikes * (bytes_per_sample * samples_per_spike * 4 +
                                                   bytes_per_timestamp * 4), num=num_spikes, endpoint=False).astype(
        int).reshape(num_spikes, 1)
    t_indices = t_start_indices

    for chan in np.arange(1, number_channels):
        t_indices = np.hstack((t_indices, t_start_indices + chan))

    t = spike_data[t_indices].reshape(num_spikes, bytes_per_timestamp)  # acquiring the time bytes
    t = np.sum(np.multiply(t, big_endian_vector), axis=1) / timebase  # converting from bytes to float values
    t_indices = None

    waveform_data = np.zeros((number_channels, num_spikes, samples_per_spike))  # (dimensions, rows, columns)

    bytes_offset = 0
    # read the t,ch1,t,ch2,t,ch3,t,ch4

    for chan in range(number_channels):  # only really care about the first time that gets written
        chan_start_indices = t_start_indices + chan * samples_per_spike + bytes_per_timestamp + bytes_per_timestamp * chan
        for spike_sample in np.arange(1, samples_per_spike):
            chan_start_indices = np.hstack((chan_start_indices, t_start_indices +
                                            chan * samples_per_spike + bytes_per_timestamp +
                                            bytes_per_timestamp * chan + spike_sample))
        waveform_data[chan][:][:] = spike_data[chan_start_indices].reshape(num_spikes, samples_per_spike).astype(
            'int8')  # acquiring the channel bytes
        waveform_data[chan][:][:][np.where(waveform_data[chan][:][:] > 127)] -= 256
        waveform_data[chan][:][:] = np.multiply(waveform_data[chan][:][:], little_endian_matrix)

    spikeparam = {'timebase': timebase, 'bytes_per_sample': bytes_per_sample, 'samples_per_spike': samples_per_spike,
                  'bytes_per_timestamp': bytes_per_timestamp, 'duration': duration, 'num_spikes': num_spikes,
                  'sample_rate': samp_rate}

    return {'t': t.reshape(num_spikes, 1), 'ch1': np.asarray(waveform_data[0][:][:]),
            'ch2': np.asarray(waveform_data[1][:][:]),
            'ch3': np.asarray(waveform_data[2][:][:]), 'ch4': np.asarray(waveform_data[3][:][:])}, spikeparam


def speed2D(x, y, t):
    '''calculates an averaged/smoothed speed'''

    N = len(x)
    v = np.zeros((N, 1))

    for index in range(1, N-1):
        v[index] = np.sqrt((x[index + 1] - x[index - 1]) ** 2 + (y[index + 1] - y[index - 1]) ** 2) / (
        t[index + 1] - t[index - 1])

    v[0] = v[1]
    v[-1] = v[-2]

    return v


def fixTimestamps(post):
    first = post[0]
    N = len(post)
    uniquePost = np.unique(post)

    if len(uniquePost) != N:
        didFix = True
        numZeros = 0
        # find the number of zeros at the end of the file

        while True:
            if post[-1 - numZeros] == 0:
                numZeros += 1
            else:
                break
        last = first + (N-1-numZeros)*0.02
        fixedPost = np.arange(first, last+0.02, 0.02)
        fixedPost = fixedPost.reshape((len(fixedPost), 1))

    else:
        didFix = False
        fixedPost = []

    return didFix, fixedPost


def remBadTrack(x, y, t, threshold):
    """function [x,y,t] = remBadTrack(x,y,t,treshold)

    % Indexes to position samples that are to be removed
   """

    remInd = []
    diffx = np.diff(x, axis=0)
    diffy = np.diff(y, axis=0)
    diffR = np.sqrt(diffx ** 2 + diffy ** 2)

    # the MATLAB works fine without NaNs, if there are Nan's just set them to threshold they will be removed later
    diffR[np.isnan(diffR)] = threshold # setting the nan values to threshold
    ind = np.where((diffR > threshold))[0]

    if len(ind) == 0:  # no bad samples to remove
        return x, y, t

    if ind[-1] == len(x):
        offset = 2
    else:
        offset = 1

    for index in range(len(ind) - offset):
        if ind[index + 1] == ind[index] + 1:
            # A single sample position jump, tracker jumps out one sample and
            # then jumps back to path on the next sample. Remove bad sample.
            remInd.append(ind[index] + 1)
        else:
            ''' Not a single jump. 2 possibilities:
             1. Tracker jumps out, and stay out at the same place for several
             samples and then jumps back.
             2. Tracker just has a small jump before path continues as normal,
             unknown reason for this. In latter case the samples are left
             untouched'''
            idx = np.where(x[ind[index] + 1:ind[index + 1] + 1 + 1] == x[ind[index] + 1])[0]
            if len(idx) == len(x[ind[index] + 1:ind[index + 1] + 1 + 1]):
                remInd.extend(
                    list(range(ind[index] + 1, ind[index + 1] + 1 + 1)))  # have that extra since range goes to end-1

    # keep_ind = [val for val in range(len(x)) if val not in remInd]
    keep_ind = np.setdiff1d(np.arange(len(x)), remInd)

    x = x[keep_ind]
    y = y[keep_ind]
    t = t[keep_ind]

    return x.reshape((len(x), 1)), y.reshape((len(y), 1)), t.reshape((len(t), 1))


def visitedBins(x, y, mapAxis):

    binWidth = mapAxis[1]-mapAxis[0]

    N = len(mapAxis)
    visited = np.zeros((N, N))

    for col in range(N):
        for row in range(N):
            px = mapAxis[col]
            py = mapAxis[row]
            distance = np.sqrt((px-x)**2 + (py-y)**2)

            if np.amin(distance) <= binWidth:
                visited[row, col] = 1

    return visited


def spikePos(ts, x, y, t, cPost, shuffleSpks, shuffleCounter=True):

    randtime = 0

    if shuffleSpks:

        # create a random sample to shuffle from -20 to 20 (not including 0)
        randsamples = np.asarray([sample_num for sample_num in range(-20, 21) if sample_num != 0])

        if shuffleCounter:
            randtime = 0
        else:
            randtime = np.random.choice(randsamples, replace=False)

            maxts = max(ts)
            ts += randtime

            if np.sign(randtime) < 0:
                for index in range(len(ts)):
                    if ts[index] < 0:
                        ts[index] = maxts + np.absolute(randtime) + ts[index]

            elif np.sign(randtime) > 0:
                for index in range(len(ts)):
                    if ts[index] > maxts:
                        ts[index] = ts[index] - maxts

            ts = np.sort(ts)
    else:
        ts = np.roll(ts, randtime)

    N = len(ts)
    spkx = np.zeros((N, 1))
    spky = np.zeros_like(spkx)
    newTs = np.zeros_like(spkx)
    count = -1 # need to subtract 1 because the python indices start at 0 and MATLABs at 1

    for index in range(N):
        tdiff = (t - ts[index])**2
        tdiff2 = (cPost-ts[index])**2
        m = np.amin(tdiff)
        ind = np.where(tdiff == m)[0]

        m2 = np.amin(tdiff2)
        #ind2 = np.where(tdiff2 == m2)[0]

        if m == m2:
            count += 1
            spkx[count] = x[ind[0]]
            spky[count] = y[ind[0]]
            newTs[count] = ts[index]

    spkx = spkx[:count + 1]
    spky = spky[:count + 1]
    newTs = newTs[:count + 1]

    return spkx, spky, newTs, randtime


def ratemap(spike_x, spike_y, posx, posy, post, h, yAxis, xAxis):
    invh = 1/h
    map = np.zeros((len(xAxis), len(yAxis)))
    pospdf = np.zeros_like(map)

    current_Y = -1
    for Y in yAxis:
        current_Y +=1
        current_X = -1
        for X in xAxis:
            current_X += 1
            map[current_Y, current_X], pospdf[current_Y, current_X] = rate_estimator(spike_x, spike_y, X, Y, invh, posx, posy, post)
    pospdf = pospdf / np.sum(np.sum(pospdf))
    return map, pospdf


def rate_estimator(spike_x, spike_y, x, y, invh, posx, posy, post):
    '''Calculate the rate for one position value.
    edge-corrected kernel density estimator'''
    conv_sum = np.sum(gaussian_kernel((spike_x-x)*invh, (spike_y-y)*invh))
    edge_corrector = np.trapz(gaussian_kernel(((posx-x)*invh),((posy-y)*invh)), post, axis=0)
    r = (conv_sum / (edge_corrector + 0.0001)) + 0.0001 # regularised firing rate for "wellbehavedness" i.e. no division by zero or log of zero
    return r, edge_corrector


def gaussian_kernel(x, y):
    '''Gaussian kernel for the rate calculation:
    % k(u) = ((2*pi)^(-length(u)/2)) * exp(u'*u)'''
    r = 0.15915494309190 * np.exp(-0.5 * (np.multiply(x,x) + np.multiply(y,y)));
    return r


def detect_peaks(x, mph=None, mpd=1, threshold=0, edge='rising',
                 kpsh=False, valley=False, show=False, ax=None):
    __author__ = "Marcos Duarte, https://github.com/demotu/BMC"
    __version__ = "1.0.4"
    __license__ = "MIT"

    """Detect peaks in data based on their amplitude and other features.

    Parameters
    ----------
    x : 1D array_like
        data.
    mph : {None, number}, optional (default = None)
        detect peaks that are greater than minimum peak height.
    mpd : positive integer, optional (default = 1)
        detect peaks that are at least separated by minimum peak distance (in
        number of data).
    threshold : positive number, optional (default = 0)
        detect peaks (valleys) that are greater (smaller) than `threshold`
        in relation to their immediate neighbors.
    edge : {None, 'rising', 'falling', 'both'}, optional (default = 'rising')
        for a flat peak, keep only the rising edge ('rising'), only the
        falling edge ('falling'), both edges ('both'), or don't detect a
        flat peak (None).
    kpsh : bool, optional (default = False)
        keep peaks with same height even if they are closer than `mpd`.
    valley : bool, optional (default = False)
        if True (1), detect valleys (local minima) instead of peaks.
    show : bool, optional (default = False)
        if True (1), plot data in matplotlib figure.
    ax : a matplotlib.axes.Axes instance, optional (default = None).

    Returns
    -------
    ind : 1D array_like
        indeces of the peaks in `x`.

    Notes
    -----
    The detection of valleys instead of peaks is performed internally by simply
    negating the data: `ind_valleys = detect_peaks(-x)`

    The function can handle NaN's

    See this IPython Notebook [1]_.

    References
    ----------
    .. [1] http://nbviewer.ipython.org/github/demotu/BMC/blob/master/notebooks/DetectPeaks.ipynb

    Examples
    --------
    >>> from detect_peaks import detect_peaks
    >>> x = np.random.randn(100)
    >>> x[60:81] = np.nan
    >>> # detect all peaks and plot data
    >>> ind = detect_peaks(x, show=True)
    >>> print(ind)

    >>> x = np.sin(2*np.pi*5*np.linspace(0, 1, 200)) + np.random.randn(200)/5
    >>> # set minimum peak height = 0 and minimum peak distance = 20
    >>> detect_peaks(x, mph=0, mpd=20, show=True)

    >>> x = [0, 1, 0, 2, 0, 3, 0, 2, 0, 1, 0]
    >>> # set minimum peak distance = 2
    >>> detect_peaks(x, mpd=2, show=True)

    >>> x = np.sin(2*np.pi*5*np.linspace(0, 1, 200)) + np.random.randn(200)/5
    >>> # detection of valleys instead of peaks
    >>> detect_peaks(x, mph=0, mpd=20, valley=True, show=True)

    >>> x = [0, 1, 1, 0, 1, 1, 0]
    >>> # detect both edges
    >>> detect_peaks(x, edge='both', show=True)

    >>> x = [-2, 1, -2, 2, 1, 1, 3, 0]
    >>> # set threshold = 2
    >>> detect_peaks(x, threshold = 2, show=True)
    """

    x = np.atleast_1d(x).astype('float64')
    if x.size < 3:
        return np.array([], dtype=int)
    if valley:
        x = -x
    # find indices of all peaks
    dx = x[1:] - x[:-1]
    # handle NaN's
    indnan = np.where(np.isnan(x))[0]
    if indnan.size:
        x[indnan] = np.inf
        dx[np.where(np.isnan(dx))[0]] = np.inf
    ine, ire, ife = np.array([[], [], []], dtype=int)
    if not edge:
        ine = np.where((np.hstack((dx, 0)) < 0) & (np.hstack((0, dx)) > 0))[0]
    else:
        if edge.lower() in ['rising', 'both']:
            ire = np.where((np.hstack((dx, 0)) <= 0) & (np.hstack((0, dx)) > 0))[0]
        if edge.lower() in ['falling', 'both']:
            ife = np.where((np.hstack((dx, 0)) < 0) & (np.hstack((0, dx)) >= 0))[0]
    ind = np.unique(np.hstack((ine, ire, ife)))
    # handle NaN's
    if ind.size and indnan.size:
        # NaN's and values close to NaN's cannot be peaks
        ind = ind[np.in1d(ind, np.unique(np.hstack((indnan, indnan - 1, indnan + 1))), invert=True)]
    # first and last values of x cannot be peaks
    if ind.size and ind[0] == 0:
        ind = ind[1:]
    if ind.size and ind[-1] == x.size - 1:
        ind = ind[:-1]
    # remove peaks < minimum peak height
    if ind.size and mph is not None:
        ind = ind[x[ind] >= mph]
    # remove peaks - neighbors < threshold
    if ind.size and threshold > 0:
        dx = np.min(np.vstack([x[ind] - x[ind - 1], x[ind] - x[ind + 1]]), axis=0)
        ind = np.delete(ind, np.where(dx < threshold)[0])
    # detect small peaks closer than minimum peak distance
    if ind.size and mpd > 1:
        ind = ind[np.argsort(x[ind])][::-1]  # sort ind by peak height
        idel = np.zeros(ind.size, dtype=bool)
        for i in range(ind.size):
            if not idel[i]:
                # keep peaks with the same height if kpsh is True
                idel = idel | (ind >= ind[i] - mpd) & (ind <= ind[i] + mpd) \
                              & (x[ind[i]] > x[ind] if kpsh else True)
                idel[i] = 0  # Keep current peak
        # remove the small peaks and sort back the indices by their occurrence
        ind = np.sort(ind[~idel])

    if show:
        if indnan.size:
            x[indnan] = np.nan
        if valley:
            x = -x
        _plot(x, mph, mpd, threshold, edge, valley, ax, ind)

    return ind


def _plot(x, mph, mpd, threshold, edge, valley, ax, ind):
    """Plot results of the detect_peaks function, see its help."""
    try:
        import matplotlib.pyplot as plt
    except ImportError:
        print('matplotlib is not available.')
    else:
        if ax is None:
            _, ax = plt.subplots(1, 1, figsize=(8, 4))

        ax.plot(x, 'b', lw=1)
        if ind.size:
            label = 'valley' if valley else 'peak'
            label = label + 's' if ind.size > 1 else label
            ax.plot(ind, x[ind], '+', mfc=None, mec='r', mew=2, ms=8,
                    label='%d %s' % (ind.size, label))
            ax.legend(loc='best', framealpha=.5, numpoints=1)
        ax.set_xlim(-.02 * x.size, x.size * 1.02 - 1)
        ymin, ymax = x[np.isfinite(x)].min(), x[np.isfinite(x)].max()
        yrange = ymax - ymin if ymax > ymin else 1
        ax.set_ylim(ymin - 0.1 * yrange, ymax + 0.1 * yrange)
        ax.set_xlabel('Data #', fontsize=14)
        ax.set_ylabel('Amplitude', fontsize=14)
        mode = 'Valley detection' if valley else 'Peak detection'
        ax.set_title("%s (mph=%s, mpd=%d, threshold=%s, edge='%s')"
                     % (mode, str(mph), mpd, str(threshold), edge))
        # plt.grid()
        plt.show()


def get_spike_color(cell_number):

    """This method will match the cell number with the color it should be RGB in Tint.

    These cells are numbered from 1-30 (there is technically a zeroth cell, but that isn't plotted"""
    spike_colors = [(1, 8, 184), (93, 249, 75), (234, 8, 9),
                         (229, 22, 239), (80, 205, 243), (27, 164, 0),
                         (251, 188, 56), (27, 143, 167), (127, 41, 116),
                         (191, 148, 23), (185, 9, 17), (231, 223, 67),
                         (144, 132, 145), (34, 236, 228), (217, 20, 145),
                         (172, 64, 80), (176, 106, 138), (199, 194, 167),
                         (216, 204, 105), (160, 204, 61), (187, 81, 88),
                         (45, 216, 122), (242, 136, 25), (50, 164, 161),
                         (249, 67, 16), (252, 232, 147), (114, 156, 238),
                         (241, 212, 179), (129, 62, 162), (235, 133, 126)]

    return spike_colors[int(cell_number)-1]


import sys, os
from cx_Freeze import setup, Executable

# Dependencies are automatically detected, but it might need fine tuning.
#build_exe_options = {"packages": ["os"], "excludes": ["tkinter"]}

# GUI applications require a different base on Windows (the default is for a
# console application).
base = None
if sys.platform == "win32":
    base = "Win32GUI"

additional_imports = ['numpy.core._methods', 'numpy.lib.format', "matplotlib.backends.backend_tkagg",
                      'scipy.spatial', 'tkinter']
                      #'TKinter', 'six']

packages = ['matplotlib', 'scipy', 'scipy.spatial', 'tkinter']


include_files = [r"C:\Python34\DLLs\tcl86t.dll",
                 r"C:\Python34\DLLs\tk86t.dll"]

os.environ['TCL_LIBRARY'] = r'C:\Python34\tcl\tcl8.6'
os.environ['TK_LIBRARY'] = r'C:\Python34\tcl\tk8.6'

setup(name="BinConverterGUI",
      version="1.0",
      description="BinConverterGUI converts the raw .bin files produced by DacqUSB to the normal Tint format." +
      "This will allow the user to select a threshold to use retroactively.",
      options={"build_exe": {'packages': packages,
                             "include_files": include_files,
                             'includes': additional_imports}},
      executables=[Executable("BinConverterGUI.py", base=base)])


import datetime, os, time
from PyQt5 import QtWidgets
import numpy as np
from BatchTINTV3.core.klusta_functions import klusta
from .Tint_Matlab import int16toint8, get_setfile_parameter
from .readBin import get_bin_data, get_raw_pos, get_channel_from_tetrode, get_active_tetrode, get_active_eeg
from .CreatePos import create_pos
from .ConvertTetrode import write_tetrode
from .CreateEEG import create_eeg, create_egf


def convert_basename(self, set_filename, threshold):
    """This function will convert the .bin file to the TINT format"""

    directory = os.path.dirname(set_filename)

    tint_basename = os.path.basename(os.path.splitext(set_filename)[0])

    bin_filename = os.path.join(directory, '%s.bin' % tint_basename)

    # the first created file of the session will be the basename for tint

    self.LogAppend.myGUI_signal.emit(
        '[%s %s]: Converting the the following session: %s!' %
        (str(datetime.datetime.now().date()),
         str(datetime.datetime.now().time())[:8], set_filename))

    # -----------------------------------Overwrite the Set File ---------------------------------------

    self.LogAppend.myGUI_signal.emit(
        '[%s %s]: Overwriting the following set file\'s duration: %s!' %
        (str(datetime.datetime.now().date()),
         str(datetime.datetime.now().time())[:8], set_filename))

    overwrite_setfile(set_filename)

    self.LogAppend.myGUI_signal.emit(
        '[%s %s]: Finished converting the following basename: %s!' %
        (str(datetime.datetime.now().date()),
         str(datetime.datetime.now().time())[:8], tint_basename))

    # --------------------------------------Read Position Data----------------------------------------

    position_filename = os.path.join(directory, tint_basename + '.pos')

    if not os.path.exists(position_filename):
        self.LogAppend.myGUI_signal.emit(
            '[%s %s]: Analyzing the following position file: %s!' %
            (str(datetime.datetime.now().date()),
             str(datetime.datetime.now().time())[:8], position_filename))

        # Fs_pos = 50  # Hz, position sampling frequency

        self.LogAppend.myGUI_signal.emit(
            '[%s %s]: Reading in the position data!' %
            (str(datetime.datetime.now().date()),
             str(datetime.datetime.now().time())[:8]))

        raw_position = get_raw_pos(bin_filename)  # vid time, x1, y1, x2, y2, numpix1, numpix2, total_pix, unused

        self.LogAppend.myGUI_signal.emit(
            '[%s %s]: Creating the .pos file!' %
            (str(datetime.datetime.now().date()),
             str(datetime.datetime.now().time())[:8]))

        create_pos(position_filename, raw_position)

    else:
        self.LogAppend.myGUI_signal.emit(
            '[%s %s]: The following position file already exists: %s!' %
            (str(datetime.datetime.now().date()),
             str(datetime.datetime.now().time())[:8], position_filename))

    # --------------------------------------Read Tetrode Data----------------------------------------

    self.LogAppend.myGUI_signal.emit(
        '[%s %s]: Converting the session one tetrode at a time!' %
        (str(datetime.datetime.now().date()),
         str(datetime.datetime.now().time())[:8]))

    Fs = get_Fs(set_filename)  # read the sampling frequency from the .set file, most like 48k

    active_tetrodes = get_active_tetrode(set_filename)

    # converts the data one tetrode at a time so we can eliminate memory errors

    pre_spike_samples = int(get_setfile_parameter('pretrigSamps', set_filename))
    post_spike_samples = int(get_setfile_parameter('spikeLockout', set_filename))
    rejstart = int(get_setfile_parameter('rejstart', set_filename))
    rejthreshtail = int(get_setfile_parameter('rejthreshtail', set_filename))
    rejthreshupper = int(get_setfile_parameter('rejthreshupper', set_filename))
    rejthreshlower = int(get_setfile_parameter('rejthreshlower', set_filename))

    for tetrode in active_tetrodes:

        tetrode = int(tetrode)
        # check if this tetrode exists already

        tetrode_filename = os.path.join(directory, '%s.%d' % (tint_basename, tetrode))
        if os.path.exists(tetrode_filename):
            self.LogAppend.myGUI_signal.emit(
                '[%s %s]: The following tetrode has already been converted, skipping: %d!' %
                (str(datetime.datetime.now().date()),
                 str(datetime.datetime.now().time())[:8], tetrode))
            continue

        self.LogAppend.myGUI_signal.emit(
            '[%s %s]: Currently Converting Data Related to Tetrode: %d!' %
            (str(datetime.datetime.now().date()),
             str(datetime.datetime.now().time())[:8], tetrode))

        tetrode_channels = get_channel_from_tetrode(tetrode)  # get the channels (from range of 1->64)

        data = get_bin_data(bin_filename, tetrode=tetrode)  # 16bit, get data associated with the tetrode

        # converting data to uV

        n_samples = data.shape[1]
        # create a time array that represents the 48kHz sampled data times
        t = np.arange(0, n_samples) / Fs  # creates a time array of the signal starting from 0 (in seconds)

        if not os.path.exists(tetrode_filename):

            # ---------------------------Find the spikes in the unit data --------------------------------------

            self.LogAppend.myGUI_signal.emit(
                '[%s %s]: Finding the spikes for the tetrode %d!' %
                (str(datetime.datetime.now().date()),
                 str(datetime.datetime.now().time())[:8], tetrode))

            tetrode_spikes = {}  # creates an empty dictionary to hold the spike times
            # for each tetrode, find the spikes

            k = 0

            # data = int16toint8(data)  # converting the data into int8

            tetrode_thresholds = []
            for channel_index, channel in enumerate(tetrode_channels):
                k += 1
                self.LogAppend.myGUI_signal.emit(
                    '[%s %s]: Finding the spikes within T%dCh%d!' %
                    (str(datetime.datetime.now().date()),
                     str(datetime.datetime.now().time())[:8], tetrode, k))

                '''Auto thresholding technique incorporated by:
                Quian Quiroga in 2014 - Unsupervised Spike Detection and Sorting with Wavelets and
                Superparamagnetic Clustering

                Thr = 4*sigma, sigma = median(abs(x)/0.6745)
                '''
                standard_deviations = float(threshold)

                sigma_n = np.median(np.divide(np.abs(data[channel_index, :]), 0.6745))
                # threshold = sigma_n / channel_max
                # threshold = standard_deviations * sigma_n
                tetrode_thresholds.append(standard_deviations * sigma_n)

            # threshold = int(17152)
            # threshold = int(16640)
            # tetrode_thresholds = [threshold, threshold, threshold, threshold]

            valid_spikes = get_spikes(data, tetrode_thresholds)

            self.LogAppend.myGUI_signal.emit(
                '[%s %s]: Number of spikes found: %d!' %
                (str(datetime.datetime.now().date()),
                 str(datetime.datetime.now().time())[:8], len(valid_spikes)))

            # threshold is done in 16 bit values, but the rejection is done in 8bit, so we convert here
            # data = int16toint8(data)  # converting the data into int8

            data = int16toint8(data)  # converting the data into int8

            tetrode_spikes = validate_spikes(self, tetrode, valid_spikes, data, t, pre_spike_samples,
                                             post_spike_samples, rejstart, rejthreshtail, rejthreshupper,
                                             rejthreshlower)

            # write the tetrode data to create the .N file
            self.LogAppend.myGUI_signal.emit(
                '[%s %s]: Writing the following tetrode file: %s!' %
                (str(datetime.datetime.now().date()),
                 str(datetime.datetime.now().time())[:8], tetrode_filename))

            write_tetrode(tetrode_filename, tetrode_spikes, Fs)
        else:
            self.LogAppend.myGUI_signal.emit(
                '[%s %s]: The following tetrode already exists, skipping: %s!' %
                (str(datetime.datetime.now().date()),
                 str(datetime.datetime.now().time())[:8], tetrode_filename))

        data = None
        tetrode_spikes = None
        valid_spikes = None

        self.LogAppend.myGUI_signal.emit(
            '[%s %s]: Finished Converting the Following Tetrode: %d!' %
            (str(datetime.datetime.now().date()),
             str(datetime.datetime.now().time())[:8], tetrode))

    ##################################################################################################
    # ----------------------------Load LFP Data then Create EEG/EGF and then Filter --------------------
    ##################################################################################################

    self.LogAppend.myGUI_signal.emit(
        '[%s %s]: Converting the EEG/EGF Files One at a Time!' %
        (str(datetime.datetime.now().date()),
         str(datetime.datetime.now().time())[:8]))

    active_eeg_channels = get_active_eeg(set_filename)
    active_eeg_channel_numbers = np.asarray(
        list(active_eeg_channels.values())) + 1  # makes the eeg channels from 1-> 64

    # check if any of the tetrode data is an EEG channel

    for channel in active_eeg_channel_numbers:

        for eeg_number, eeg_chan_value in sorted(active_eeg_channels.items()):
            if eeg_chan_value == channel - 1:
                break

        if eeg_number == 1:
            eeg_str = ''
        else:
            eeg_str = str(eeg_number)

        if eeg_number == 1:
            eeg_filename = os.path.join(directory, tint_basename + '.eeg')
            egf_filename = os.path.join(directory, tint_basename + '.egf')
        else:
            eeg_filename = os.path.join(directory, tint_basename + '.eeg%d' % (eeg_number))
            egf_filename = os.path.join(directory, tint_basename + '.egf%d' % (eeg_number))

        if os.path.exists(eeg_filename):

            EEG = np.array([])
            self.LogAppend.myGUI_signal.emit(
                '[%s %s]: The following EEG file has already been created, skipping: %s!' %
                (str(datetime.datetime.now().date()),
                 str(datetime.datetime.now().time())[:8], eeg_filename))
        else:
            self.LogAppend.myGUI_signal.emit(
                '[%s %s]: Creating the following EEG file: .eeg%s!' %
                (str(datetime.datetime.now().date()),
                 str(datetime.datetime.now().time())[:8], eeg_str))

            # load data
            EEG = get_bin_data(bin_filename, channels=[channel])

            create_eeg(eeg_filename, EEG, Fs, DC_Blocker=self.dc_blocker.isChecked())
            # EEG = None

        if os.path.exists(egf_filename):
            self.LogAppend.myGUI_signal.emit(
                '[%s %s]: The following EGF file has already been created, skipping: %s!' %
                (str(datetime.datetime.now().date()),
                 str(datetime.datetime.now().time())[:8], eeg_filename))
        else:

            if len(EEG) != 0:
                # then the EEG has already been loaded in
                create_egf(egf_filename, EEG, Fs, DC_Blocker=self.dc_blocker.isChecked())
                EEG = None
            else:
                self.LogAppend.myGUI_signal.emit(
                    '[%s %s]: Creating the following EGF file: .egf%s!' %
                    (str(datetime.datetime.now().date()),
                     str(datetime.datetime.now().time())[:8], eeg_str))

                # then the EEG hasn't been read in (EEG was already created), read the data
                EGF = get_bin_data(bin_filename, channels=[channel])

                create_egf(egf_filename, EGF, Fs, DC_Blocker=self.dc_blocker.isChecked())
                EGF = None

            # session_parameters['SamplesPerSpike'] = 50
            # session_parameters['timebase'] = 96000
        EEG = None

    self.LogAppend.myGUI_signal.emit(
        '[%s %s]: Finished converting the following: %s!' %
        (str(datetime.datetime.now().date()),
         str(datetime.datetime.now().time())[:8], tint_basename))


def matching_ind(haystack, needle):
    idx = np.searchsorted(haystack, needle)
    mask = idx < haystack.size
    mask[mask] = haystack[idx[mask]] == needle[mask]
    idx = idx[mask]
    return idx


def batchtint(main_window, directory):

    check_chosen_settings_directory(main_window)

    if main_window.choice == QtWidgets.QMessageBox.Abort:
        main_window.LogAppend.myGUI_signal.emit(
            '[%s %s]: Batch-Tint of the following directory has been aborted: %s!' % (str(datetime.datetime.now().date()),
                                                                 str(datetime.datetime.now().time())[
                                                                 :8], directory))
        main_window.StopConversion()
        return

    check_valid_settings_directory(main_window)

    if main_window.choice == QtWidgets.QMessageBox.Abort:
        main_window.LogAppend.myGUI_signal.emit(
            '[%s %s]: Batch-Tint of the following directory has been aborted: %s!' % (str(datetime.datetime.now().date()),
                                                                 str(datetime.datetime.now().time())[
                                                                 :8], directory))
        main_window.StopConversion()
        return

    main_window.settings_fname = main_window.parameters['tintsettings']

    klusta_ready = True

    if klusta_ready:
        main_window.LogAppend.myGUI_signal.emit(
            '[%s %s]: Analyzing the following directory: %s!' % (str(datetime.datetime.now().date()),
                                                                 str(datetime.datetime.now().time())[
                                                                 :8], directory))
        # ------------- find all subdirectories within directory -------------------------------------

        sub_directories = [d for d in os.listdir(directory) if
                           os.path.isdir(os.path.join(directory, d)) and d not in ['Processed', 'Converted']]

        if not sub_directories:

            # message that shows how many files were found
            main_window.LogAppend.myGUI_signal.emit(
                '[%s %s]: There are no files to analyze in this directory!' % (str(datetime.datetime.now().date()),
                                                                               str(datetime.datetime.now().time())[
                                                                               :8]))
            return

        else:
            # message that shows how many files were found
            main_window.LogAppend.myGUI_signal.emit(
                '[%s %s]: Found %d files in the directory!' % (str(datetime.datetime.now().date()),
                                                               str(datetime.datetime.now().time())[
                                                               :8], len(sub_directories)))

        # ----------- cycle through each file and find the tetrode files ------------------------------------------

        for sub_directory in sub_directories:  # finding all the folders within the directory
            try:
                dir_new = os.path.join(directory, sub_directory)  # sets a new filepath for the directory
                f_list = os.listdir(dir_new)  # finds the files within that directory
                set_file = [file for file in f_list if '.set' in file]  # finds the set file

                if not set_file:  # if there is no set file it will return as an empty list
                    # message saying no .set file
                    main_window.LogAppend.myGUI_signal.emit(
                        '[%s %s]: The following folder contains no \'.set\' file: %s' % (
                            str(datetime.datetime.now().date()),
                            str(datetime.datetime.now().time())[
                            :8], str(sub_directory)))
                    continue
                # runs the function that will perform the klusta'ing
                klusta(main_window, sub_directory, directory)

            except NotADirectoryError:
                # if the file is not a directory it prints this message
                main_window.LogAppend.myGUI_signal.emit(
                    '[%s %s]: %s is not a directory!' % (
                        str(datetime.datetime.now().date()),
                        str(datetime.datetime.now().time())[
                        :8], str(sub_directory)))
                continue


def check_chosen_settings_directory(main_window):
    # checks if the settings are appropriate to run analysis
    if 'Choose the Batch-Tint' in main_window.parameters['tintsettings']:
        main_window.choice = None
        main_window.directory_chosen = False
        main_window.LogError.myGUI_signal.emit('NoTintSettings')

        while main_window.choice is None:
            time.sleep(0.5)

        if main_window.choice == QtWidgets.QMessageBox.Ok:

            while not main_window.directory_chosen:
                time.sleep(0.1)

            main_window.parameters['tintsettings'] = os.path.join(
                main_window.batchtintsettings_edit.text(), 'settings.json')


def check_valid_settings_directory(main_window):

    while True:
        if 'Choose the Batch-Tint' in main_window.parameters['tintsettings']:
            check_chosen_settings_directory()

            if main_window.choice == QtWidgets.QMessageBox.Abort:
                return

            if 'Choose the Batch-Tint' in main_window.parameters['tintsettings']:
                main_window.choice = None
                main_window.directory_chosen = False
                main_window.LogError.myGUI_signal.emit('DefaultTintSettings')

                while main_window.choice is None:
                    time.sleep(0.5)

                if main_window.choice == QtWidgets.QMessageBox.Yes:
                    main_window.SETTINGS_DIR = main_window.BATCH_TINT_DIR
                    main_window.parameters['tintsettings'] = os.path.join(main_window.SETTINGS_DIR, 'settings.json')
                else:
                    return
        else:
            if not os.path.exists(os.path.join(main_window.parameters['tintsettings'])):
                main_window.choice = None
                main_window.directory_chosen = False
                main_window.LogError.myGUI_signal.emit('InvalidTintSettings')

                while main_window.choice is None:
                    time.sleep(0.5)

                if main_window.choice == QtWidgets.QMessageBox.Yes:

                    while not main_window.directory_chosen:
                        time.sleep(0.5)

                    main_window.parameters['tintsettings'] = os.path.join(
                        main_window.batchtintsettings_edit.text(), 'settings.json')

                elif main_window.choice == QtWidgets.QMessageBox.Default:
                    main_window.SETTINGS_DIR = main_window.BATCH_TINT_DIR
                    main_window.parameters['tintsettings'] = os.path.join(main_window.SETTINGS_DIR, 'settings.json')
                else:
                    return
            else:
                return


def get_Fs(set_filename):
    fs_str = 'rawRate'

    with open(set_filename) as f:
        for line in f:

            if fs_str in line:
                _, Fs = line.split(' ')

                try:
                    return int(Fs)
                except:
                    return float(Fs)
    return


def find_consec(data):
    '''finds the consecutive numbers and outputs as a list'''
    consecutive_values = []  # a list for the output
    current_consecutive = [data[0]]

    if len(data) == 1:
        return [[data[0]]]

    for index in range(1, len(data)):

        if data[index] == data[index - 1] + 1:
            current_consecutive.append(data[index])

            if index == len(data) - 1:
                consecutive_values.append(current_consecutive)

        else:
            consecutive_values.append(current_consecutive)
            current_consecutive = [data[index]]

            if index == len(data) - 1:
                consecutive_values.append(current_consecutive)
    return consecutive_values


def get_spikes(data, threshold):
    all_spikes = np.array([])

    for i, channel_data in enumerate(data):
        spike_indices = np.where(channel_data >= threshold[i])[0]

        if len(spike_indices) == 0:
            continue

        spike_indices = find_consec(spike_indices)

        spike_indices = np.asarray([value[0] for value in spike_indices])

        if len(all_spikes) == 0:
            # this is the first iteration of the tetrode, no need to sort
            unadded_spikes = spike_indices
        else:
            idx = matching_ind(all_spikes, spike_indices)
            if len(idx) == 0:
                unadded_spikes = spike_indices
            else:
                unadded_spikes = np.setdiff1d(spike_indices, all_spikes[idx])

        if len(all_spikes) != 0:
            all_spikes = np.sort(np.concatenate((all_spikes, unadded_spikes)))
            unadded_spikes = None
        else:
            all_spikes = np.array(unadded_spikes)

    return all_spikes


def validate_spikes(self, tetrode, spikes, data, t, pre_spike_samples=10, post_spike_samples=40, rejstart=30,
                    rejthreshtail=43, rejthreshupper=100, rejthreshlower=-100):
    latest_spike = None

    spike_count = 0
    percentage_values = [int(value) for value in np.rint(np.linspace(0, len(spikes), num=21)).tolist()]

    n_max = data.shape[1]

    tetrode_spikes = {}

    for spike in sorted(spikes):
        # iterate through each spike and validate to ensure no spikes occur at the same time or within the
        # refractory period

        spike_count += 1

        if spike_count in percentage_values:
            self.LogAppend.myGUI_signal.emit(
                '[%s %s]: %d%% completed adding spikes from T%d!' %
                (str(datetime.datetime.now().date()),
                 str(datetime.datetime.now().time())[:8], int(np.rint(100 * spike_count / len(spikes))),
                 tetrode))

        if spike - pre_spike_samples + 1 < 0:
            continue

        elif spike + post_spike_samples >= n_max:
            continue

        if latest_spike is not None:
            if spike != latest_spike:
                if spike in spike_refractory:
                    # ensures no overlapping spikes
                    continue
        else:
            pass

        latest_spike = spike
        spike_refractory = list(np.arange(spike + 1, spike + post_spike_samples + 1))

        # spike_time = t[int(spike)]
        spike_time = t[int(spike)]

        # waveform_indices = np.where((t>=spike_time-250/1e6) & (t<=spike_time+850/1e6))[0]  # too slow
        waveform_indices = np.arange(spike - pre_spike_samples + 1, spike + post_spike_samples + 1).astype(int)

        # spike_t = t[waveform_indices] - spike_time  # making the times from -200 us to 800 us

        # spike_waveform = np.zeros((len(tetrode_channels), 50))

        spike_waveform = data[:, waveform_indices]

        spike_time = spike_time * 96000  # multiply it by the timebase to get the frame count

        spike_waveform = np.rint(spike_waveform)

        # artifact rejection

        if sum(spike_waveform[:, rejstart:].flatten() > rejthreshtail) > 0:
            # this is 33% above baseline (0)
            continue

        # check if the first sample is well above or well below baseline
        elif sum(spike_waveform[:, 0].flatten() > rejthreshupper) > 0:
            # the first sample is >100
            continue

        elif sum(spike_waveform[:, 0].flatten() < rejthreshlower) > 0:
            # or < -100
            continue

        tetrode_spikes[spike_time] = spike_waveform

        # latest_spike = spike
        # spike_refractory = list(np.arange(spike + 1, spike + post_spike_samples + 1))

    return tetrode_spikes


def overwrite_setfile(set_filename):
    with open(set_filename, 'r+') as f:

        header = ''
        footer = ''

        header_values = True

        for line in f:

            if header_values:
                if 'duration' in line:
                    line = line.strip()
                    duration = int(np.ceil(float(line.split(' ')[-1])))
                    header += 'duration %d\n' % duration
                    continue
                header += line

            else:
                footer += line

    with open(set_filename, 'w') as f:

        f.writelines([header, footer])




from scipy import signal, fftpack
# import matplotlib.pyplot as plt
import numpy as np


def notch_filt(data, Fs, band=10, freq=60, ripple=1, order=2, filter_type='butter', analog_filt=False,
               showresponse=0):
    '''# Required input defintions are as follows;
    # time:   Time between samples
    # band:   The bandwidth around the centerline freqency that you wish to filter
    # freq:   The centerline frequency to be filtered
    # ripple: The maximum passband ripple that is allowed in db
    # order:  The filter order.  For FIR notch filters this is best set to 2 or 3,
    #         IIR filters are best suited for high values of order.  This algorithm
    #         is hard coded to FIR filters
    # filter_type: 'butter', 'bessel', 'cheby1', 'cheby2', 'ellip'
    # data:         the data to be filtered'''

    cutoff = freq
    nyq = Fs / 2.0
    low = freq - band / 2.0
    high = freq + band / 2.0
    low = low / nyq
    high = high / nyq
    b, a = signal.iirfilter(order, [low, high], rp=ripple, btype='bandstop', analog=analog_filt, ftype=filter_type)

    filtered_data = np.array([])

    if len(data) != 0:
        if len(data.shape) > 1:  # lfilter is one dimensional so we need to perform for loop on multi-dimensional array
            # filtered_data = np.zeros((data.shape[0], data.shape[1]))
            filtered_data = signal.filtfilt(b, a, data, axis=1)
            # for channel_num in range(0, data.shape[0]):
            # filtered_data[channel_num,:] = signal.lfilter(b, a, data[channel_num,:])
            #   filtered_data[channel_num, :] = signal.filtfilt(b, a, data[channel_num, :])
        else:
            # filtered_data = signal.lfilter(b, a, data)
            filtered_data = signal.filtfilt(b, a, data)

    FType = ''

    '''
    if showresponse == 1:
        if filter_type == 'butter':
            FType = 'Butterworth'
        elif filter_type == 'cheby1':
            FType = 'Chebyshev I'
        elif filter_type == 'cheby2':
            FType = 'Chebyshev II'
        elif filter_type == 'ellip':
            FType = 'Cauer/Elliptic'
        elif filter_type == 'bessel':
            FType = 'Bessel/Thomson'

        if analog_filt == 1:
            mode = 'Analog'
        else:
            mode = 'Digital'

        if analog_filt is False:
            w, h = signal.freqz(b, a, worN=8000)  # returns the requency response h, and the normalized angular
            # frequencies w in radians/sample
            # w (radians/sample) * Fs (samples/sec) * (1 cycle/2pi*radians) = Hz
            f = Fs * w / (2 * np.pi)  # Hz
        else:
            w, h = signal.freqs(b, a, worN=8000)  # returns the requency response h, and the angular frequencies
            # w in radians/sec
            # w (radians/sec) * (1 cycle/2pi*radians) = Hz
            f = w / (2 * np.pi)  # Hz

        plt.figure(figsize=(20, 15))
        plt.subplot(211)
        plt.semilogx(f, np.abs(h), 'b')
        plt.xscale('log')
        plt.title('%s Filter Frequency Response (%s)' % (FType, mode))
        plt.xlabel('Frequency(Hz)')
        plt.ylabel('Gain [V/V]')
        plt.margins(0, 0.1)
        plt.grid(which='both', axis='both')
        plt.axvline(cutoff, color='green')
    '''
    return filtered_data


def get_a_b(bandtype, Fs, Wp, Ws, order=3, Rp=3, As=60, analog_val=False, filttype='butter', automatic=0):

    stop_amp = 1.5
    stop_amp2 = 1.4

    if not analog_val:  # need to convert the Ws and Wp to have the units of pi radians/sample
        # this is for digital filters
        if bandtype in ['low', 'high']:
            Wp = Wp / (Fs / 2)  # converting to fraction of nyquist frequency

            Ws = Wp * stop_amp

        elif bandtype == 'band':
            Wp = Wp / (Fs / 2)  # converting to fraction of nyquist frequency
            Wp2 = Wp / stop_amp2

            Ws = Ws / (Fs / 2)  # converting to fraction of nyquist frequency
            Ws2 = Ws * stop_amp2

    else:  # need to convert the Ws and Wp to have the units of radians/sec
        # this is for analog filters
        if bandtype in ['low', 'high']:
            Wp = 2 * np.pi * Wp

            Ws = Wp * stop_amp

        elif bandtype == 'band':
            Wp = 2 * np.pi * Wp
            Wp2 = Wp / stop_amp2

            Ws = 2 * np.pi * Ws
            Ws2 = Ws * stop_amp2

    if automatic == 1:
        if bandtype in ['low', 'high']:
            b, a = signal.iirdesign(wp=Wp, ws=Ws, gpass=Rp, gstop=As, analog=analog_val, ftype=filttype)
        elif bandtype == 'band':
            b, a = signal.iirdesign(wp=[Wp, Ws], ws=[Wp2, Ws2], gpass=Rp, gstop=As, analog=analog_val, ftype=filttype)
    else:
        if bandtype in ['low', 'high']:
            if filttype == 'cheby1' or 'cheby2' or 'ellip':
                b, a = signal.iirfilter(order, Wp, rp=Rp, rs=As, btype=bandtype, analog=analog_val, ftype=filttype)
            else:
                b, a = signal.iirfilter(order, Wp, btype=bandtype, analog=analog_val, ftype=filttype)
        elif bandtype == 'band':
            if filttype == 'cheby1' or 'cheby2' or 'ellip':
                b, a = signal.iirfilter(order, [Wp, Ws], rp=Rp, rs=As, btype=bandtype, analog=analog_val,
                                        ftype=filttype)
            else:
                b, a = signal.iirfilter(order, [Wp, Ws], btype=bandtype, analog=analog_val, ftype=filttype)

    return b, a


def iirfilt(bandtype, data, Fs, Wp, Ws=[], order=3, analog_val=False, automatic=0, Rp=3, As=60, filttype='butter',
            showresponse=0):
    '''Designs butterworth filter:
    Data is the data that you want filtered
    Fs is the sampling frequency (in Hz)
    Ws and Wp are stop and pass frequencies respectively (in Hz)

    Passband (Wp) : This is the frequency range which we desire to let the signal through with minimal attenuation.
    Stopband (Ws) : This is the frequency range which the signal should be attenuated.

    Digital: Ws is the normalized stop frequency where 1 is the nyquist freq (pi radians/sample in digital)
             Wp is the normalized pass frequency

    Analog: Ws is the stop frequency in (rads/sec)
            Wp is the pass frequency in (rads/sec)

    Analog is false as default, automatic being one has Python select the order for you. pass_atten is the minimal attenuation
    the pass band, stop_atten is the minimal attenuation in the stop band. Fs is the sample frequency of the signal in Hz.

    Rp = 0.1      # passband maximum loss (gpass)
    As = 60 stoppand min attenuation (gstop)


    filttype : str, optional
        The type of IIR filter to design:
        Butterworth : ‘butter’
        Chebyshev I : ‘cheby1’
        Chebyshev II : ‘cheby2’
        Cauer/elliptic: ‘ellip’
        Bessel/Thomson: ‘bessel’

    bandtype : {‘bandpass’, ‘lowpass’, ‘highpass’, ‘bandstop’}, optional
    '''

    cutoff = Wp

    if Ws != []:
        cutoff2 = Ws

    b, a = get_a_b(bandtype, Fs, Wp, Ws, order=order, Rp=Rp, As=As, analog_val=analog_val, filttype=filttype,
                   automatic=automatic)
    if len(data) != 0:
        if len(data.shape) > 1:

            filtered_data = np.zeros((data.shape[0], data.shape[1]))
            filtered_data = signal.filtfilt(b, a, data, axis=1)
        else:
            filtered_data = signal.filtfilt(b, a, data)
    '''
    if showresponse == 1:  # set to 1 if you want to visualize the frequency response of the filter
        
        if filttype == 'butter':
            FType = 'Butterworth'
        elif filttype == 'cheby1':
            FType = 'Chebyshev I'
        elif filttype == 'cheby2':
            FType = 'Chebyshev II'
        elif filttype == 'ellip':
            FType = 'Cauer/Elliptic'
        elif filttype == 'bessel':
            FType = 'Bessel/Thomson'

        if analog_val:
            mode = 'Analog'
        else:
            mode = 'Digital'

        if not analog_val:
            w, h = signal.freqz(b, a, worN=8000)  # returns the requency response h, and the normalized angular
            # frequencies w in radians/sample
            # w (radians/sample) * Fs (samples/sec) * (1 cycle/2pi*radians) = Hz
            f = Fs * w / (2 * np.pi)  # Hz
        else:
            w, h = signal.freqs(b, a, worN=8000)  # returns the requency response h,
            # and the angular frequencies w in radians/sec
            # w (radians/sec) * (1 cycle/2pi*radians) = Hz
            f = w / (2 * np.pi)  # Hz

        plt.figure(figsize=(10, 5))
        plt.semilogx(f, np.abs(h), 'b')
        plt.xscale('log')

        if 'cutoff2' in locals():
            plt.title('%s Bandpass Filter Frequency Response (Order = %s, Wp=%s (Hz), Ws =%s (Hz))'
                      % (FType, order, cutoff, cutoff2))
        else:
            plt.title('%s Lowpass Filter Frequency Response (Order = %s, Wp=%s (Hz))'
                      % (FType, order, cutoff))

        plt.xlabel('Frequency(Hz)')
        plt.ylabel('Gain [V/V]')
        plt.margins(0, 0.1)
        plt.grid(which='both', axis='both')
        plt.axvline(cutoff, color='green')
        if 'cutoff2' in locals():
            plt.axvline(cutoff2, color='green')
            # plt.plot(cutoff, 0.5*np.sqrt(2), 'ko') # cutoff frequency
        plt.show()
    '''

    if len(data) != 0:
        return filtered_data


def custom_cheby1(data, Fs, N, Rp, Wp, Ws=None, filtresponse='bandpass', analog_value=False, showresponse=0):
    nyquist = Fs/2

    if filtresponse == 'bandpass':
        Wn = [Wp/nyquist, Ws/nyquist]
    else:
        Wn = [Wp/nyquist]

    b, a = signal.cheby1(N, Rp, Wn, 'bandpass', analog=analog_value)

    if data != []:
        if len(data.shape) > 1:
            #print('Filtering multidimensional array!')
            filtered_data = np.zeros((data.shape[0], data.shape[1]))
            filtered_data = signal.filtfilt(b, a, data, axis=1)
            #for channel_num in range(0, data.shape[0]):
            #    # filtered_data[channel_num,:] = signal.lfilter(b, a, data[channel_num,:])
            #    filtered_data[channel_num, :] = signal.filtfilt(b, a, data[channel_num, :])
        else:
            # filtered_data = signal.lfilter(b, a, data)
            filtered_data = signal.filtfilt(b, a, data)

    '''
    if showresponse == 1:
        FType = 'Chebyshev I'
        mode = 'Digital'

        w, h = signal.freqz(b, a, worN=8000)  # returns the requency response h, and the normalized angular
        # frequencies w in radians/sample
        # w (radians/sample) * Fs (samples/sec) * (1 cycle/2pi*radians) = Hz
        f = Fs * w / (2 * np.pi)  # Hz

        plt.figure(figsize=(10, 5))
        # plt.subplot(211)
        plt.semilogx(f, np.abs(h), 'b')
        plt.xscale('log')

        if Ws is not None:
            plt.title('%s Bandpass Filter Frequency Response (Order = %s, Wp=%s (Hz), Ws =%s (Hz))'
                      % (FType, N, Wp, Ws))
        else:
            plt.title('%s Lowpass Filter Frequency Response (Order = %s, Wp=%s (Hz))'
                      % (FType, N, Wp))

        plt.xlabel('Frequency(Hz)')
        plt.ylabel('Gain [V/V]')
        plt.margins(0, 0.1)
        plt.grid(which='both', axis='both')
        plt.axvline(Wp, color='green')
        if Ws is not None:
            plt.axvline(Ws, color='green')
            # plt.plot(Ws, 0.5*np.sqrt(2), 'ko') # cutoff frequency
        plt.show()
    '''

    return filtered_data


def dcblock(data, fc, fs=None, analog_val=False, showresponse=0, self=None):

    """This method will return the filter coefficients for a DC Blocker Filter"""

    if fs is None:
        Fc = fc

    else:
        Fc = 2 * fc / fs

    p = (np.sqrt(3) - 2 * np.sin(np.pi * Fc)) / (np.sin(np.pi * Fc) +
                                                 np.sqrt(3) * np.cos(np.pi * Fc));

    b = np.array([1, -1])
    a = np.array([1, -p])

    if len(data) != 0:
        if len(data.shape) > 1:

            # filtered_data = np.zeros((data.shape[0], data.shape[1]))
            filtered_data = signal.filtfilt(b, a, data, axis=1)

        else:
            # filtered_data = signal.lfilter(b, a, data)
            filtered_data = signal.filtfilt(b, a, data)
    '''
    if showresponse == 1:  # set to 1 if you want to visualize the frequency response of the filter
        if self is not None:
            self.PlotResponse(a, b, fc, fs, analog_val)
    '''
    if len(data) != 0:
        return filtered_data


'''
def PlotResponse(a, b, fc, fs, analog_val):
    if analog_val:
        mode = 'Analog'
    else:
        mode = 'Digital'

    if not analog_val:
        w, h = signal.freqz(b, a, worN=8000)  # returns the requency response h, and the normalized angular
        # frequencies w in radians/sample
        # w (radians/sample) * Fs (samples/sec) * (1 cycle/2pi*radians) = Hz
        f = fs * w / (2 * np.pi)  # Hz
    else:
        w, h = signal.freqs(b, a, worN=8000)  # returns the requency response h,
        # and the angular frequencies w in radians/sec
        # w (radians/sec) * (1 cycle/2pi*radians) = Hz
        f = w / (2 * np.pi)  # Hz

    fig = plt.figure(figsize=(10, 5))
    ax = fig.add_subplot(111)
    # plt.subplot(211)
    # ax.plot(f, np.abs(h), 'b')
    # ax.set_xlim([0,1])
    ax.semilogx(f, np.abs(h), 'b')
    ax.set_xscale('log')
    # ax.set_xlim([-1000, ax.get_xlim()[1]])

    plt.title("DC Block Filter")

    ax.set_xlabel('Frequency(Hz)')
    ax.set_ylabel('Gain [V/V]')
    ax.margins(0, 0.1)
    ax.grid(which='both', axis='both')
    ax.axvline(fc, color='green')
    plt.show()



def fft_plot(Fs, Y):
    """Takes the Sample Frequency: Fs(Hz), the numer of samples, N, and the data values (Y),
    and performs a Fast Fourier Transformation to observe the signal in the frequency domain"""

    N = len(Y)
    k = np.arange(N)

    T = N / Fs

    frq = k / T  # two sides frequency range
    frq = frq[range(np.int(N / 2))]  # one side frequency range

    FFT = fftpack.fft(Y)  # fft computing and normalization
    fig, ax = plt.subplots(1)
    ax.plot(frq, 2.0 / N * np.abs(FFT[0:int(N / 2)]), 'b')

    ax.set_title('FFT Plot')
    ax.set_xlabel('Frequency(Hz)')
    ax.set_ylabel('Amplitude')
    ax.margins(0, 0.1)
    ax.grid()

    return fig, ax
'''

def FastFourier(Fs, Y):
    '''Takes the Sample Frequency: Fs(Hz), the numer of samples, N, and the data values (Y),
    and performs a Fast Fourier Transformation to observe the signal in the frequency domain'''

    N = len(Y)
    k = np.arange(N)

    T = N / Fs

    frq = k / T  # two sides frequency range
    frq = frq[range(np.int(N / 2))]  # one side frequency range

    FFT = fftpack.fft(Y)  # fft computing and normalization

    FFT_norm = 2.0 / N * np.abs(FFT[0:int(N / 2)])

    return frq, FFT_norm



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

pkgs = setuptools.find_packages()
print('found these packages:', pkgs)

pkg_name = "BinConverter"

setuptools.setup(
    name=pkg_name,
    version="1.0.3",
    author="Geoffrey Barrett",
    author_email="geoffrey.m.barrett@gmail.com",
    description="BinConverter - GUI that will allow the user to convert the raw (.bin) files from Axona's " +
                "Tint software to their normal Tint format",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/HussainiLab/BinConverter.git",
    packages=pkgs,
    install_requires=
    [
        'PyQt5',
        'BatchTINTV3',
        'numpy',
        'scipy'
    ],
    package_data={'BinConverter': ['img/*.png']},
    classifiers=[
        "Programming Language :: Python :: 3.7 ",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3) ",
        "Operating System :: OS Independent",
    ],
)



from __future__ import absolute_import, division, print_function

import sys
import os
sys.path.append(os.path.dirname(__file__))

from .core.AddSessions import *
from .core.conversion_utils import *
from .core.ConversionFunctions import *
from .core.ConvertTetrode import *
from .core.CreateEEG import *
from .core.CreatePos import *
from .core.filtering import *
from .core.readBin import *
from .core.Tint_Matlab import *
from .core.utils import *

__all__ = ['core', 'BinConverterGUI']

# __path__ = __import__('pkgutil').extend_path(__path__, __name__)





import sys, os
from PyQt5 import QtWidgets, QtGui, QtCore
import time


_author_ = "Geoffrey Barrett"  # defines myself as the author

Large_Font = ("Arial", 11)  # defines two fonts for different purposes (might not be used
Small_Font = ("Arial", 8)

project_name = 'BinConverter'


@QtCore.pyqtSlot()
def raise_w(new_window, old_window):
    """ raise the current window"""

    if "Settings_Window" in str(new_window):
        new_window.raise_window()
        old_window.hide()

    elif "Settings_Window" in str(old_window):
        new_window.raise_()
        new_window.show()

        old_window.backbtn_function()
        old_window.hide()
    else:
        new_window.raise_()
        new_window.show()
        time.sleep(0.1)
        old_window.hide()


class Communicate(QtCore.QObject):
    '''A custom pyqtsignal so that errors and popups can be called from the threads
    to the main window'''
    myGUI_signal = QtCore.pyqtSignal(str)
    myGUI_signal_str = myGUI_signal
    myGUI_signal_QTreeWidgetItem = QtCore.pyqtSignal(QtWidgets.QTreeWidgetItem)


class Worker(QtCore.QObject):
    '''This worker object will act to ensure that the QThreads are not within the Main Thread'''
    # def __init__(self, main_window, thread):
    def __init__(self, function, *args, **kwargs):
        '''takes in a function, and the arguments and keyword arguments for that function'''
        super(Worker, self).__init__()
        self.function = function
        self.args = args
        self.kwargs = kwargs
        self.start.connect(self.run)

    start = QtCore.pyqtSignal(str)

    @QtCore.pyqtSlot()
    def run(self):
        self.function(*self.args, **self.kwargs)


def background(self):  # defines the background for each window
    """providing the background info for each window"""
    # Acquiring information about geometry
    project_dir = os.path.dirname(os.path.abspath("__file__"))

    if os.path.basename(project_dir) != project_name:
        project_dir = os.path.dirname(sys.argv[0])

    # defining the directory filepaths
    self.PROJECT_DIR = project_dir  # project directory

    self.IMG_DIR = os.path.join(self.PROJECT_DIR, 'img')  # image directory
    self.CORE_DIR = os.path.join(self.PROJECT_DIR, 'core')  # core directory
    self.SETTINGS_DIR = os.path.join(self.PROJECT_DIR, 'settings')  # settings directory
    if not os.path.exists(self.SETTINGS_DIR):
        os.mkdir(self.SETTINGS_DIR)

    self.BATCH_TINT_DIR = os.path.join(self.PROJECT_DIR, 'BatchTint')
    self.setWindowIcon(QtGui.QIcon(os.path.join(self.IMG_DIR, 'GEBA_Logo.png')))  # declaring the icon image
    self.deskW, self.deskH = QtWidgets.QDesktopWidget().availableGeometry().getRect()[2:]  # gets the window resolution
    # self.setWindowState(QtCore.Qt.WindowMaximized) # will maximize the GUI
    self.setGeometry(0, 0, self.deskW/2, self.deskH/1.75)

    self.settings_fname = os.path.join(self.SETTINGS_DIR, 'settings.json')
    self.directory_settings = os.path.join(self.SETTINGS_DIR, 'directory.json')

    QtWidgets.QApplication.setStyle(QtWidgets.QStyleFactory.create('Cleanlooks'))


def center(self):
    """centers the window on the screen"""
    frameGm = self.frameGeometry()
    screen = QtWidgets.QApplication.desktop().screenNumber(QtWidgets.QApplication.desktop().cursor().pos())
    centerPoint = QtWidgets.QApplication.desktop().screenGeometry(screen).center()
    frameGm.moveCenter(centerPoint)
    self.move(frameGm.topLeft())


def find_consec(data):
    '''finds the consecutive numbers and outputs as a list'''
    consecutive_values = []  # a list for the output
    current_consecutive = [data[0]]

    if len(data) == 1:
        return [[data[0]]]

    for index in range(1, len(data)):

        if data[index] == data[index - 1] + 1:
            current_consecutive.append(data[index])

            if index == len(data) - 1:
                consecutive_values.append(current_consecutive)

        else:
            consecutive_values.append(current_consecutive)
            current_consecutive = [data[index]]

            if index == len(data) - 1:
                consecutive_values.append(current_consecutive)
    return consecutive_values

